from PyQt5.QtCore import Qt, QPoint
from PyQt5.QtGui import QMouseEvent, QIcon
from PyQt5.QtWidgets import QLabel, QApplication, QAction, QToolBar, QToolButton

from gui_tools import RS
from patchbay.tools_widgets import PatchbayToolsWidget
from patchbay.tool_bar import PatchbayToolBar
from patchbay.patchbay_manager import PatchbayManager

_translate = QApplication.translate


class RayToolBar(PatchbayToolBar):
    def __init__(self, parent):
        super().__init__(parent)
        self.force_main_actions_icons_only : bool = RS.settings.value(
            'tool_bar/icons_only', False, type=bool)
    
    def set_patchbay_manager(self, patchbay_manager: PatchbayManager):
        super().set_patchbay_manager(patchbay_manager)
        self._set_main_actions_icon_only(
            self.force_main_actions_icons_only)
    
    def _set_main_actions_icon_only(self, yesno: bool):
        if yesno:
            self.setToolButtonStyle(Qt.ToolButtonIconOnly)
        else:
            self.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)

        self.force_main_actions_icons_only = yesno
    
    def mousePressEvent(self, event: QMouseEvent):
        child_widget = self.childAt(event.pos())
        QToolBar.mousePressEvent(self, event)

        if not (event.button() == Qt.RightButton
                and (child_widget is None
                     or isinstance(child_widget, (QLabel, PatchbayToolsWidget)))):
            return

        context_actions = self._make_context_actions()
        menu = self._make_context_menu(context_actions)
        
        act_text_with_icons = QAction(QIcon.fromTheme('format-text-direction-symbolic'),
                                      _translate('gui_tool_bar', 'Text with session actions'))
        act_text_with_icons.setCheckable(True)
        act_text_with_icons.setChecked(not self.force_main_actions_icons_only)
        menu.addSeparator()
        menu.addAction(act_text_with_icons)
        
        # execute the menu, exit if no action
        point = event.screenPos().toPoint()
        point.setY(self.mapToGlobal(QPoint(0, self.height())).y())
        selected_act = menu.exec(point)
        if selected_act is None:
            return
        
        if selected_act is act_text_with_icons:
            self._set_main_actions_icon_only(not act_text_with_icons.isChecked())
            return

        for key, act in context_actions.items():
            if act is selected_act:
                if act.isChecked():
                    self._displayed_widgets |= key
                else:
                    self._displayed_widgets &= ~key

        self._change_visibility()



#!/usr/bin/python3 -u

#libs
import signal
import sys
import time

from PyQt5.QtWidgets import QApplication
from PyQt5.QtGui import QIcon, QFontDatabase
from PyQt5.QtCore import QLocale, QTranslator, QTimer, QLibraryInfo

#local imports
from gui_tools import ArgParser, CommandLineArgs, init_gui_tools, get_code_root
from gui_server_thread import GuiServerThread
from gui_session import SignaledSession
import ray


def signal_handler(sig, frame):
    if sig in (signal.SIGINT, signal.SIGTERM):
        if session.daemon_manager.launched_before:
            if (CommandLineArgs.under_nsm
                    and session.server_status != ray.ServerStatus.OFF):
                session.main_win.terminate_request = True

                l_server = GuiServerThread.instance()
                if l_server:
                    l_server.abort_session()
            else:
                session.daemon_manager.stop()
            return

        session.main_win.terminate_request = True
        session.daemon_manager.stop()

if __name__ == '__main__':
    # set Qt Application
    app = QApplication(sys.argv)
    app.setApplicationName(ray.APP_TITLE)
    app.setApplicationVersion(ray.VERSION)
    app.setOrganizationName(ray.APP_TITLE)
    app.setWindowIcon(QIcon(
        f':main_icon/scalable/{ray.APP_TITLE.lower()}.svg'))
    app.setQuitOnLastWindowClosed(False)
    app.setDesktopFileName(ray.APP_TITLE.lower())

    # with some themes (GNOME fedora 34)
    # QGroupBox are not really visible
    app.setStyleSheet("QGroupBox{background-color: #15888888}")

    ### Translation process
    locale = QLocale.system().name()

    app_translator = QTranslator()
    if app_translator.load(QLocale(), ray.APP_TITLE.lower(),
                           '_', str(get_code_root() / 'locale')):
        app.installTranslator(app_translator)

    patchbay_translator = QTranslator()
    if patchbay_translator.load(
            QLocale(), 'patchbay',
            '_', str(get_code_root() / 'HoustonPatchbay' / 'locale')):
        app.installTranslator(patchbay_translator)

    sys_translator = QTranslator()
    path_sys_translations = QLibraryInfo.location(QLibraryInfo.TranslationsPath)
    if sys_translator.load(QLocale(), 'qt', '_', path_sys_translations):
        app.installTranslator(sys_translator)

    QFontDatabase.addApplicationFont(":/fonts/Ubuntu-R.ttf")
    QFontDatabase.addApplicationFont(":/fonts/Ubuntu-C.ttf")

    # get arguments
    parser = ArgParser()

    init_gui_tools()

    # Add raysession/src/bin to $PATH
    # to can use raysession after make, without install
    ray.add_self_bin_to_path()

    #connect signals
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    #needed for signals SIGINT, SIGTERM
    timer = QTimer()
    timer.start(200)
    timer.timeout.connect(lambda: None)

    #build session
    server = GuiServerThread()
    session = SignaledSession()

    app.exec()

    # TODO find something better, sometimes program never ends without.
    time.sleep(0.002)

    server.stop()
    session.quit()
    del session
    del app



#!/usr/bin/python3 -u

from dataclasses import dataclass
from enum import IntEnum
import os
import signal
import sys
from typing import Optional
import warnings

import osc_server
import threading
import time

import jacklib
from jacklib.helpers import c_char_p_p_to_list
from jacklib.enums import JackOptions, JackPortFlags, JackMetadata

# check ALSA LIB
try:
    from pyalsa.alsaseq import SEQ_LIB_VERSION_STR
    ALSA_VERSION_LIST = [int(num) for num in SEQ_LIB_VERSION_STR.split('.')]
    assert ALSA_VERSION_LIST >= [1, 2, 4]
    ALSA_LIB_OK = True
    from alsa_manager import AlsaManager
except:
    ALSA_LIB_OK = False
    

PORT_TYPE_NULL = 0
PORT_TYPE_AUDIO = 1
PORT_TYPE_MIDI = 2

EXISTENCE_PATH = '/tmp/RaySession/patchbay_daemons/'



# Define a context manager to suppress stdout and stderr.
class suppress_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in 
    Python, i.e. will suppress all print, even if the print originates in a 
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).      

    '''
    def __init__(self):
        # Open a pair of null files
        self.null_fds =  [os.open(os.devnull,os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0],1)
        os.dup2(self.null_fds[1],2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0],1)
        os.dup2(self.save_fds[1],2)
        # Close all file descriptors
        for fd in self.null_fds + self.save_fds:
            os.close(fd)
            

@dataclass
class TransportPosition:
    frame: int
    rolling: bool
    valid_bbt: bool
    bar: int
    beat: int
    tick: int
    beats_per_minutes: float


class TransportWanted(IntEnum):
    NO = 0         # do not send any transport info
    STATE_ONLY = 1 # send info only when play/pause changed
    FULL = 2       # send all Transport infos


class JackPort:
    id = 0
    name = ''
    type = PORT_TYPE_NULL
    flags = 0
    alias_1 = ''
    alias_2 = ''
    order = None
    uuid = 0
    
    def __init__(self, port_name:str, jack_client, port_ptr=None):
        # In some cases, port could has just been renamed
        # then, jacklib.port_by_name() fail.
        # that is why, port_ptr can be sent as argument here
        self.name = port_name
        if port_ptr is None:
            port_ptr = jacklib.port_by_name(jack_client, port_name)
        self.flags = jacklib.port_flags(port_ptr)
        self.uuid = jacklib.port_uuid(port_ptr)

        port_type_str = jacklib.port_type(port_ptr)
        if port_type_str == jacklib.JACK_DEFAULT_AUDIO_TYPE:
            self.type = PORT_TYPE_AUDIO
        elif port_type_str == jacklib.JACK_DEFAULT_MIDI_TYPE:
            self.type = PORT_TYPE_MIDI
            
        order_prop = jacklib.get_property(
            self.uuid, JackMetadata.ORDER)

        ret, alias_1, alias_2 = jacklib.port_get_aliases(port_ptr)
        if ret:
            self.alias_1 = alias_1
            self.alias_2 = alias_2

    def __lt__(self, other: 'JackPort'):
        return self.uuid < other.uuid


class MainObject:
    port_list = list[JackPort]()
    connection_list = list[tuple[str]]()
    metadata_list = list[dict]()
    client_list = list[dict]()
    client_names_queue = list[str]()
    jack_running = False
    osc_server = None
    alsa_mng: Optional['AlsaManager'] = None
    terminate = False
    jack_client = None
    samplerate = 48000
    buffer_size = 1024
    
    dsp_wanted = True
    transport_wanted = TransportWanted.FULL
    
    def __init__(self, daemon_port: str, gui_url: str):
        self._daemon_port = daemon_port
        self.ALSA_LIB_OK = ALSA_LIB_OK
        self.last_sent_dsp_load = 0
        self.max_dsp_since_last_sent = 0.00
        self._waiting_jack_client_open = True
        self.last_transport_pos = TransportPosition(0, False, False, 0, 0, 0, 0.0)

        self.osc_server = osc_server.OscJackPatch(self)
        self.osc_server.set_tmp_gui_url(gui_url)
        self.write_existence_file()
        self.start_jack_client()
        
        if ALSA_LIB_OK:
            self.alsa_mng = AlsaManager(self)
            self.alsa_mng.add_all_ports()
    
    @staticmethod
    def get_metadata_value_str(prop: jacklib.Property) -> str:
        value = prop.value
        if isinstance(value, bytes):
            return value.decode()
        elif isinstance(value, str):
            return value
        else:
            try:
                value = str(value)
            except:
                return ''
        return value
    
    def write_existence_file(self):
        if not os.path.isdir(EXISTENCE_PATH):
            os.makedirs(EXISTENCE_PATH)
        
        try:
            file = open(EXISTENCE_PATH + self._daemon_port, 'w')
        except PermissionError:
            sys.stderr.write(
                'ray-patchbay_to_osc: Error, no permission for existence file\n')
            sys.exit(1)

        contents = 'pid:%i\n' % os.getpid()
        contents += 'port:%i\n' % self.osc_server.port

        file.write(contents)
        file.close()
    
    def remove_existence_file(self):
        if not os.path.exists(EXISTENCE_PATH + self._daemon_port):
            return 

        try:
            os.remove(EXISTENCE_PATH + self._daemon_port)
        except PermissionError:
            sys.stderr.write(
                'ray-patchbay_to_osc: Error, unable to remove %s\n'
                % EXISTENCE_PATH + self._daemon_port)
    
    @classmethod
    def signal_handler(cls, sig: int, frame):
        if sig in (signal.SIGINT, signal.SIGTERM):
            cls.terminate = True
            
    def eat_client_names_queue(self):
        while self.client_names_queue:
            client_name = self.client_names_queue.pop(0)
            b_uuid = jacklib.get_uuid_for_client_name(self.jack_client, client_name)

            # convert bytes uuid to int
            uuid = 0
            if isinstance(b_uuid, bytes):
                str_uuid = b_uuid.decode()
                if str_uuid.isdigit():
                    uuid = int(str_uuid)

            if not uuid:
                continue

            for client_dict in self.client_list:
                if client_dict['name'] == client_name:
                    client_dict['uuid'] = uuid
                    break
            else:
                self.client_list.append({'name': client_name, 'uuid': uuid})
    
    def add_gui(self, gui_url: str):
        self.osc_server.add_gui(gui_url)
    
    def check_jack_client_responding(self):
        for i in range(100): # JACK has 5s to answer
            time.sleep(0.050)

            if not self._waiting_jack_client_open:
                break
        else:
            # server never answer
            self.osc_server.send_server_lose()
            self.remove_existence_file()
            
            # JACK is not responding at all
            # probably it is started but totally bugged
            # finally kill this program from system
            os.kill(os.getpid(), signal.SIGKILL)
    
    def refresh(self):
        if self.jack_running:
            self.get_all_ports_and_connections()
            self.osc_server.server_restarted()
            
        if self.alsa_mng is not None:
            self.alsa_mng.add_all_ports()
    
    def remember_dsp_load(self):
        self.max_dsp_since_last_sent = max(
            self.max_dsp_since_last_sent,
            jacklib.cpu_load(self.jack_client))
        
    def send_dsp_load(self):
        current_dsp = int(self.max_dsp_since_last_sent + 0.5)
        if current_dsp != self.last_sent_dsp_load:
            self.osc_server.send_dsp_load(current_dsp)
            self.last_sent_dsp_load = current_dsp
        self.max_dsp_since_last_sent = 0.00
    
    def set_transport_wanted(self, transport_wanted: int):
        try:
            self.transport_wanted = TransportWanted(transport_wanted)
        except:
            self.transport_wanted = TransportWanted.FULL

    def _send_transport_pos(self):
        if not self.jack_running:
            return
        
        pos = jacklib.jack_position_t()
        pos.valid = 0

        state = jacklib.transport_query(self.jack_client, jacklib.pointer(pos))
        
        if (self.transport_wanted is TransportWanted.STATE_ONLY
                and bool(state) == self.last_transport_pos.rolling):
            return

        transport_position = TransportPosition(
            int(pos.frame),
            bool(state),
            bool(pos.valid & jacklib.JackPositionBits.POSITION_BBT),
            int(pos.bar),
            int(pos.beat),
            int(pos.tick),
            float(pos.beats_per_minute))
        
        if transport_position == self.last_transport_pos:
            return
        
        self.last_transport_pos = transport_position
        self.osc_server.send_transport_position(transport_position)
    
    def connect_ports(self, port_out_name: str, port_in_name: str,
                      disconnect=False):
        if self.alsa_mng is not None and port_out_name.startswith(':ALSA_OUT:'):
            self.alsa_mng.connect_ports(
                port_out_name, port_in_name, disconnect=disconnect)
            return

        if disconnect:
            jacklib.disconnect(self.jack_client, port_out_name, port_in_name)
        else:
            jacklib.connect(self.jack_client, port_out_name, port_in_name)
    
    def start_loop(self):
        n = 0

        while True:
            self.osc_server.recv(50)
            
            if self.is_terminate():
                break

            if self.jack_running:
                if n % 4 == 0:
                    self.remember_dsp_load()
                    if self.dsp_wanted and n % 20 == 0:
                        self.send_dsp_load()
                
                self.eat_client_names_queue()
                if self.transport_wanted is not TransportWanted.NO:
                    self._send_transport_pos()

            else:
                if n % 10 == 0:
                    self.start_jack_client()
            n += 1
            
            # for faster modulos
            if n == 20:
                n = 0
                
    def exit(self):
        if self.jack_running:
            jacklib.deactivate(self.jack_client)
            jacklib.client_close(self.jack_client)

        if self.alsa_mng is not None:
            self.alsa_mng.stop_events_loop()
            del self.alsa_mng

        self.remove_existence_file()
        del self.osc_server
    
    def start_jack_client(self):
        self._waiting_jack_client_open = True
        
        # Sometimes JACK never registers the client
        # and never answers. This thread will allow to exit
        # if JACK didn't answer 5 seconds after register ask
        jack_waiter_thread = threading.Thread(
            target=self.check_jack_client_responding)
        jack_waiter_thread.start()

        with suppress_stdout_stderr():
            self.jack_client = jacklib.client_open(
                "ray-patch_to_osc",
                JackOptions.NO_START_SERVER | JackOptions.SESSION_ID,
                None)

        self._waiting_jack_client_open = False
        ## Some problems happens to the JACK server sometimes without it
        #time.sleep(0.030)

        jack_waiter_thread.join()

        if self.jack_client:
            self.jack_running = True
            self.set_registrations()
            self.get_all_ports_and_connections()
            self.samplerate = jacklib.get_sample_rate(self.jack_client)
            self.buffer_size = jacklib.get_buffer_size(self.jack_client)
            self.osc_server.server_restarted()
        else:
            self.jack_running = False

    def is_terminate(self)->bool:
        if self.terminate or self.osc_server.is_terminate():
            return True
        
        return False
    
    def set_registrations(self):
        if not self.jack_client:
            return
        
        jacklib.set_client_registration_callback(
            self.jack_client, self.jack_client_registration_callback, None)
        jacklib.set_port_registration_callback(
            self.jack_client, self.jack_port_registration_callback, None)
        jacklib.set_port_connect_callback(
            self.jack_client, self.jack_port_connect_callback, None)
        jacklib.set_port_rename_callback(
            self.jack_client, self.jack_port_rename_callback, None)
        jacklib.set_xrun_callback(
            self.jack_client, self.jack_xrun_callback, None)
        jacklib.set_buffer_size_callback(
            self.jack_client, self.jack_buffer_size_callback, None)
        jacklib.set_sample_rate_callback(
            self.jack_client, self.jack_sample_rate_callback, None)
        jacklib.set_property_change_callback(
            self.jack_client, self.jack_properties_change_callback, None)
        jacklib.on_shutdown(
            self.jack_client, self.jack_shutdown_callback, None)
        jacklib.activate(self.jack_client)
    
    def get_all_ports_and_connections(self):
        self.port_list.clear()
        self.connection_list.clear()
        self.metadata_list.clear()

        #get all currents Jack ports and connections
        port_name_list = c_char_p_p_to_list(
            jacklib.get_ports(self.jack_client, "", "", 0))
        
        client_names = []
        
        for port_name in port_name_list:
            port_ptr = jacklib.port_by_name(self.jack_client, port_name)
            jport = JackPort(port_name, self.jack_client)
            self.port_list.append(jport)
            
            client_name = port_name.partition(':')[0]
            if not client_name in client_names:
                client_names.append(client_name)

            # get port metadatas
            for key in (JackMetadata.CONNECTED,
                        JackMetadata.ORDER,
                        JackMetadata.PORT_GROUP,
                        JackMetadata.PRETTY_NAME,
                        JackMetadata.SIGNAL_TYPE):
                prop = jacklib.get_property(jport.uuid, key)
                if prop is None:
                    continue

                value = self.get_metadata_value_str(prop)
                self.metadata_list.append(
                    {'uuid': jport.uuid,
                     'key': key,
                     'value': value})

            if jport.flags & JackPortFlags.IS_INPUT:
                continue

            port_ptr = jacklib.port_by_name(self.jack_client, jport.name)
            
            # this port is output, list its connections
            port_connection_names = tuple(
                jacklib.port_get_all_connections(self.jack_client, port_ptr))

            for port_con_name in port_connection_names:
                self.connection_list.append((jport.name, port_con_name))
        
        for client_name in client_names:
            uuid = jacklib.get_uuid_for_client_name(self.jack_client, client_name)
            if not uuid:
                continue

            self.client_list.append({'name': client_name, 'uuid': int(uuid)})
            
            # we only look for icon_name now, but in the future other client
            # metadatas could be enabled
            for key in (JackMetadata.ICON_NAME,):
                prop = jacklib.get_property(int(uuid), JackMetadata.ICON_NAME)
                if prop is None:
                    continue
                value = self.get_metadata_value_str(prop)
                self.metadata_list.append(
                    {'uuid': int(uuid),
                    'key': key,
                    'value': value})
    
    def jack_shutdown_callback(self, arg=None)->int:
        self.jack_running = False
        self.port_list.clear()
        self.connection_list.clear()
        self.osc_server.server_stopped()
        return 0

    def jack_xrun_callback(self, arg=None)->int:
        self.osc_server.send_one_xrun()
        return 0

    def jack_sample_rate_callback(self, samplerate, arg=None) -> int:
        self.samplerate = samplerate
        self.osc_server.send_samplerate()
        return 0

    def jack_buffer_size_callback(self, buffer_size, arg=None) -> int:
        self.buffer_size = buffer_size
        self.osc_server.send_buffersize()
        return 0

    def jack_client_registration_callback(self, client_name: bytes,
                                          register: int, arg=None) -> int:
        client_name = client_name.decode()
        self.client_names_queue.append(client_name)
        return 0
        
    def jack_port_registration_callback(self, port_id: int, register: bool,
                                        arg=None) -> int:
        if not self.jack_client:
            return 0
        
        port_ptr = jacklib.port_by_id(self.jack_client, port_id)
        port_name = jacklib.port_name(port_ptr)
        
        if register:
            jport = JackPort(port_name, self.jack_client, port_ptr)
            self.port_list.append(jport)
            self.osc_server.port_added(
                jport.name, jport.type, jport.flags, jport.uuid)
        else:
            for jport in self.port_list:
                if jport.name == port_name:
                    self.port_list.remove(jport)
                    self.osc_server.port_removed(jport.name)
                    break
        return 0
    
    def jack_port_rename_callback(self, port_id: int, old_name: bytes,
                                  new_name: bytes, arg=None)->int:
        for jport in self.port_list:
            if jport.name == str(old_name.decode()):
                ex_name = jport.name
                jport.name = str(new_name.decode())
                self.osc_server.port_renamed(ex_name, jport.name)
                break
        return 0
    
    def jack_port_connect_callback(self, port_id_A: int, port_id_B: int,
                                   connect_yesno: bool, arg=None)->int:
        port_ptr_A = jacklib.port_by_id(self.jack_client, port_id_A)
        port_ptr_B = jacklib.port_by_id(self.jack_client, port_id_B)

        port_str_A = jacklib.port_name(port_ptr_A)
        port_str_B = jacklib.port_name(port_ptr_B)

        connection = (port_str_A, port_str_B)

        if connect_yesno:
            self.connection_list.append(connection)
            self.osc_server.connection_added(connection)
        elif connection in self.connection_list:
            self.connection_list.remove(connection)
            self.osc_server.connection_removed(connection)

        return 0

    def jack_properties_change_callback(self, uuid: int, name: bytes,
                                        type_: int, arg=None)->int:
        if name is not None:
            name = name.decode()
        
        value = ''

        if name and type_ != jacklib.JackPropertyChange.DELETED:
            prop = jacklib.get_property(uuid, name)
            if prop is None:
                return 0

            value = self.get_metadata_value_str(prop)
        
        for metadata in self.metadata_list:
            if metadata['uuid'] == uuid and metadata['key'] == name:
                metadata['value'] = value
                break
        else:
            self.metadata_list.append(
                {'uuid': uuid, 'key': name, 'value': value})

        self.osc_server.metadata_updated(uuid, name, value)

        return 0
    
    def set_buffer_size(self, buffer_size: int):
        jacklib.set_buffer_size(self.jack_client, buffer_size)

    def set_metadata(self, uuid: int, key: str, value: str):
        jacklib.set_property(uuid, key, value, 'text/plain', jacklib.ENCODING)

    def transport_play(self, play: bool):
        if play:
            jacklib.transport_start(self.jack_client)
        else:
            jacklib.transport_stop(self.jack_client)
            
    def transport_stop(self):
        jacklib.transport_stop(self.jack_client)
        jacklib.transport_locate(self.jack_client, 0)
        
    def transport_relocate(self, frame: int):
        jacklib.transport_locate(self.jack_client, frame)


def main_process():
    args = sys.argv.copy()
    daemon_port = ''
    gui_url = ''

    if args:
        this_exec = args.pop(0)

    if args:
        daemon_port = args.pop(0)
    
    if args:
        gui_url = args[0]
    
    main_object = MainObject(daemon_port, gui_url)

    for gui_url in args:
        main_object.add_gui(gui_url)
    
    main_object.start_loop()
    main_object.exit()


if __name__ == '__main__':
    # prevent deprecation warnings python messages
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    
    signal.signal(signal.SIGINT, MainObject.signal_handler)
    signal.signal(signal.SIGTERM, MainObject.signal_handler)
    
    main_process()


import sys
import pytest

import jacklib
from jacklib.helpers import get_jack_status_error_string

@pytest.fixture
def jack_client():
    status = jacklib.jack_status_t()
    client = jacklib.client_open("pyjacklib", jacklib.JackOptions.NO_START_SERVER, status)

    if status.value:
        err = get_jack_status_error_string(status)

        if status.value & jacklib.JackStatus.NAME_NOT_UNIQUE:
            print(f"Non-fatal JACK status: {err}", file=sys.stderr)
        elif status.value & jacklib.JackStatus.SERVER_STARTED:
            # Should not happen, since we use the JackOptions.NO_START_SERVER option
            print(f"Unexpected JACK status: {err}", file=sys.stderr)
        else:
            raise OSError(f"Error creating JACK client: {err}")

    yield client

    jacklib.client_close(client)



"""Helper functions for extra jacklib functionality."""

# Copyright (C) 2010-2020 Filipe Coelho <falktx@falktx.com>
#               2016-2022 Christopher Arndt <info@chrisarndt.de>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# For a full copy of the GNU General Public License see the file COPYING.md.

from ctypes import pointer, c_char_p
from typing import Iterator
from . import api as jacklib
from .api import JackStatus


def get_jack_status_error_string(cStatus):
    """Get JACK error status as string."""

    status = cStatus.value

    if status == 0x0:
        return ""

    try:
        status = JackStatus(status)
    except:
        return ""

    error_string = list[str]()

    if status is JackStatus.FAILURE:
        # Only include this generic message if no other error status is set
        error_string.append("Overall operation failed")
    if status & JackStatus.INVALID_OPTION:
        error_string.append("The operation contained an invalid or unsupported option")
    if status & JackStatus.NAME_NOT_UNIQUE:
        error_string.append("The desired client name was not unique")
    if status & JackStatus.SERVER_STARTED:
        error_string.append("The JACK server was started as a result of this operation")
    if status & JackStatus.SERVER_FAILED:
        error_string.append("Unable to connect to the JACK server")
    if status & JackStatus.SERVER_ERROR:
        error_string.append("Communication error with the JACK server")
    if status & JackStatus.NO_SUCH_CLIENT:
        error_string.append("Requested client does not exist")
    if status & JackStatus.LOAD_FAILURE:
        error_string.append("Unable to load internal client")
    if status & JackStatus.INIT_FAILURE:
        error_string.append("Unable to initialize client")
    if status & JackStatus.SHM_FAILURE:
        error_string.append("Unable to access shared memory")
    if status & JackStatus.VERSION_ERROR:
        error_string.append("Client's protocol version does not match")
    if status & JackStatus.BACKEND_ERROR:
        error_string.append("Backend Error")
    if status & JackStatus.CLIENT_ZOMBIE:
        error_string.append("Client is being shutdown against its will")

    return ";\n".join(error_string) + "."


def c_char_p_p_to_list(c_char_p_p: 'pointer[c_char_p]',
                       encoding=jacklib.ENCODING,
                       errors="ignore") -> list[str]:
    """Convert C char** -> Python list of strings."""
    i = 0
    ret_list = list[str]()

    if not c_char_p_p:
        return ret_list

    while True:
        new_char_p = c_char_p_p[i]
        if not new_char_p:
            break

        ret_list.append(new_char_p.decode(encoding=encoding, errors=errors))
        i += 1

    jacklib.free(c_char_p_p)
    return ret_list


def iterate_c_char_p_p(c_char_p_p: 'pointer[c_char_p]',
                       encoding=jacklib.ENCODING,
                       errors="ignore") -> Iterator[str]:
    i = 0
    
    if not c_char_p:
        return
    
    while True:
        new_char_p = c_char_p_p[i]
        if not new_char_p:
            break
        
        yield new_char_p.decode(encoding=encoding, errors=errors)
        i += 1
        
    jacklib.free(c_char_p_p)


def voidptr2str(void_p) -> str:
    """Convert C void* -> string."""
    char_p = jacklib.cast(void_p, jacklib.c_char_p)
    string = str(char_p.value, encoding="utf-8")
    return string


def translate_audio_port_buffer(void_p):
    """Convert C void* -> jack_default_audio_sample_t*."""

    return jacklib.cast(void_p, jacklib.POINTER(jacklib.jack_default_audio_sample_t))


def translate_midi_event_buffer(void_p, size):
    """Convert a JACK MIDI buffer into a Python tuple of 0-4 elements."""

    if not void_p:
        return ()
    elif size == 1:
        return (void_p[0],)
    elif size == 2:
        return (void_p[0], void_p[1])
    elif size == 3:
        return (void_p[0], void_p[1], void_p[2])
    elif size == 4:
        return (void_p[0], void_p[1], void_p[2], void_p[3])
    else:
        return ()



from dataclasses import dataclass
from enum import IntFlag, IntEnum, auto
from typing import TYPE_CHECKING, Any, Union

from .patchcanvas import (patchcanvas, PortMode, PortType, BoxType,
                          BoxLayoutMode, BoxSplitMode, PortSubType)

if TYPE_CHECKING:
    from .patchbay_manager import PatchbayManager

# Port Flags as defined by JACK
class JackPortFlag(IntFlag):
    IS_INPUT = 0x01
    IS_OUTPUT = 0x02
    IS_PHYSICAL = 0x04
    CAN_MONITOR = 0x08
    IS_TERMINAL = 0x10
    IS_CONTROL_VOLTAGE = 0x100


class GroupPosFlag(IntFlag):
    # used in some config files,
    # it explains why some numbers are missing.
    SPLITTED = 0x04
    WRAPPED_INPUT = 0x10
    WRAPPED_OUTPUT = 0x20
    HAS_BEEN_SPLITTED = 0x40


class PortTypesViewFlag(IntFlag):
    NONE = 0x00
    AUDIO = 0x01
    MIDI = 0x02
    CV = 0x04
    VIDEO = 0x08
    ALSA = 0x10
    ALL = AUDIO | MIDI | CV | VIDEO | ALSA


@dataclass
class TransportPosition:
    frame: int
    rolling: bool
    valid_bbt: bool
    bar: int
    beat: int
    tick: int
    beats_per_minutes: float


class TransportViewMode(IntEnum):
    HOURS_MINUTES_SECONDS = 0
    BEAT_BAR_TICK = 1
    FRAMES = 2


class ToolDisplayed(IntFlag):
    PORT_TYPES_VIEW = auto()
    TRANSPORT_CLOCK = auto()
    TRANSPORT_PLAY_STOP = auto()
    TRANSPORT_TEMPO = auto()
    ZOOM_SLIDER = auto()
    BUFFER_SIZE = auto()
    SAMPLERATE = auto()
    LATENCY = auto()
    XRUNS = auto()
    DSP_LOAD = auto()
    ALL = (PORT_TYPES_VIEW
           | TRANSPORT_CLOCK
           | TRANSPORT_PLAY_STOP
           | TRANSPORT_TEMPO
           | ZOOM_SLIDER
           | BUFFER_SIZE
           | SAMPLERATE
           | LATENCY
           | XRUNS
           | DSP_LOAD)
    
    def to_save_string(self) -> str:
        ''' returns a string containing all flags names
            separated with pipe symbol.'''
        all_strs = list[str]()
        
        for flag in ToolDisplayed:
            if flag is ToolDisplayed.ALL:
                continue

            if self & flag:
                all_strs.append(flag.name)
            else:
                all_strs.append('~' + flag.name)
        
        return '|'.join(all_strs)
    
    def filtered_by_string(self, string: str) -> 'ToolDisplayed':
        '''returns another ToolDisplayed with value filtered
           by string where string contains flags names separated with pipe symbol
           as given by to_save_string method.'''
        return_td = ToolDisplayed(self.value)
        
        for disp_str in string.split('|'):
            delete = False
            if disp_str.startswith('~'):
                delete = True
                disp_str = disp_str[1:]

            if disp_str in ToolDisplayed._member_names_:
                if delete:
                    return_td &= ~ToolDisplayed[disp_str]
                else:
                    return_td |= ToolDisplayed[disp_str]

        return return_td


class GroupPos:
    port_types_view: PortTypesViewFlag = PortTypesViewFlag.NONE
    group_name: str = ""
    null_zone: str = ""
    in_zone: str = ""
    out_zone: str = ""
    null_xy: tuple[int, int]
    in_xy: tuple[int, int]
    out_xy: tuple[int, int]
    flags: int = 0
    layout_modes: dict[PortMode, BoxLayoutMode]
    fully_set: bool = True
    
    def __init__(self):
        self.null_xy = (0, 0)
        self.in_xy = (0, 0)
        self.out_xy = (0, 0)
        self.layout_modes = dict[PortMode, BoxLayoutMode]()
    
    @staticmethod
    def _is_point(value: Any) -> bool:
        if not isinstance(value, (list, tuple)):
            return False
        
        if len(value) != 2:
            return False
        
        for v in value:
            if not isinstance(v, int):
                return False
        return True

    @staticmethod
    def from_serialized_dict(src: dict[str, Any]) -> 'GroupPos':
        port_types_view = src['port_types_view']
        group_name = src['group_name']
        null_zone = src['null_zone']
        out_zone = src['out_zone']
        in_zone = src['in_zone']
        null_xy = src['null_xy']
        in_xy = src['in_xy']
        out_xy = src['out_xy']
        flags = src['flags']
        layout_modes = src['layout_modes']
        
        gpos = GroupPos()
        
        if isinstance(port_types_view, int):
            gpos.port_types_view = PortTypesViewFlag(
                port_types_view & PortTypesViewFlag.ALL)
        if isinstance(group_name, str):
            gpos.group_name = group_name
        if isinstance(null_zone, str):
            gpos.null_zone = null_zone
        if isinstance(in_zone, str):
            gpos.in_zone = in_zone
        if isinstance(out_zone, str):
            gpos.out_zone = out_zone
        if GroupPos._is_point(null_xy):
            gpos.null_xy = tuple(null_xy)
        if GroupPos._is_point(in_xy):
            gpos.in_xy = tuple(in_xy)
        if GroupPos._is_point(out_xy):
            gpos.out_xy = tuple(out_xy)
        if isinstance(flags, int):
            try:
                gpos.flags = GroupPosFlag(flags)
            except:
                pass

        if isinstance(layout_modes, dict):
            for key, value in layout_modes.items():
                try:
                    gpos.layout_modes[PortMode(int(key))] = BoxLayoutMode(int(value))
                except:
                    pass
        
        return gpos

    def copy(self) -> 'GroupPos':
        group_pos = GroupPos()
        group_pos.__dict__ = self.__dict__.copy()
        return group_pos

    def eat(self, other: 'GroupPos'):
        self.__dict__ = other.__dict__.copy()

    def as_serializable_dict(self):
        return {'port_types_view': int(self.port_types_view),
                'group_name': self.group_name,
                'null_zone': self.null_zone,
                'in_zone': self.in_zone,
                'out_zone': self.out_zone,
                'null_xy': self.null_xy,
                'in_xy': self.in_xy,
                'out_xy': self.out_xy,
                'flags': self.flags,
                'layout_modes': self.layout_modes}
        
    def set_layout_mode(self, port_mode: PortMode, layout_mode: BoxLayoutMode):
        self.layout_modes[port_mode] = layout_mode

    def get_layout_mode(self, port_mode: PortMode) -> BoxLayoutMode:
        if port_mode in self.layout_modes.keys():
            return self.layout_modes[port_mode]
        return BoxLayoutMode.AUTO


class PortgroupMem:
    group_name: str = ""
    port_type: PortType = PortType.NULL
    port_mode: PortMode = PortMode.NULL
    port_names: list[str]
    above_metadatas: bool = False
    
    def __init__(self):
        self.port_names = list[str]()

    @staticmethod
    def from_serialized_dict(src: dict[str, Any]) -> 'PortgroupMem':
        pg_mem = PortgroupMem()

        try:
            pg_mem.group_name = str(src['group_name'])
            pg_mem.port_type = PortType(src['port_type'])
            pg_mem.port_mode = PortMode(src['port_mode'])
            pg_mem.port_names = [str(a) for a in src['port_names']]
            pg_mem.above_metadatas = bool(src['above_metadatas'])
        except:
            pass

        return pg_mem

    def has_a_common_port_with(self, other: 'PortgroupMem') -> bool:
        if (self.port_type is not other.port_type
                or self.port_mode is not other.port_mode
                or self.group_name != other.group_name):
            return False
        
        for port_name in self.port_names:
            if port_name in other.port_names:
                return True
        
        return False
    
    def as_serializable_dict(self) -> dict[str, Any]:
        return {
            'group_name': self.group_name,
            'port_type': self.port_type,
            'port_mode': self.port_mode,
            'port_names': self.port_names,
            'above_metadatas': self.above_metadatas
        }


class Connection:
    def __init__(self, manager: 'PatchbayManager', connection_id: int,
                 port_out: 'Port', port_in: 'Port'):
        self.manager = manager
        self.connection_id = connection_id
        self.port_out = port_out
        self.port_in = port_in
        self.in_canvas = False

    def port_type(self) -> PortType:
        return self.port_out.type

    def full_type(self) -> tuple[PortType, PortSubType]:
        port_out_type, port_out_subtype = self.port_out.full_type()
        port_in_type, port_in_subtype = self.port_in.full_type()
        return (port_out_type, port_out_subtype | port_in_subtype)
        
    def shown_in_port_types_view(self, port_types_view: PortTypesViewFlag) -> bool:
        if self.port_out.type is PortType.MIDI_JACK:
            return bool(port_types_view & PortTypesViewFlag.MIDI)
        
        if (self.port_out.type is PortType.AUDIO_JACK
                and self.port_in.type is PortType.AUDIO_JACK):
            if (self.port_out.subtype is PortSubType.CV
                    and self.port_in.subtype is PortSubType.CV): 
                return bool(port_types_view & PortTypesViewFlag.CV)
            if (self.port_out.subtype is PortSubType.REGULAR
                    and self.port_in.subtype is PortSubType.REGULAR):
                return bool(port_types_view & PortTypesViewFlag.AUDIO)
        
        return False

    def add_to_canvas(self):
        if self.manager.very_fast_operation:
            return

        if self.in_canvas:
            return

        if not self.manager.port_type_shown(self.full_type()):
            return

        self.in_canvas = True

        patchcanvas.connect_ports(
            self.connection_id,
            self.port_out.group_id, self.port_out.port_id,
            self.port_in.group_id, self.port_in.port_id)

    def remove_from_canvas(self):
        if self.manager.very_fast_operation:
            return

        if not self.in_canvas:
            return

        patchcanvas.disconnect_ports(self.connection_id)
        self.in_canvas = False

    def semi_hide(self, yesno: bool):
        if not self.in_canvas:
            return
        
        patchcanvas.semi_hide_connection(
            self.connection_id, yesno)
    
    def set_in_front(self):
        if not self.in_canvas:
            return
        
        patchcanvas.set_connection_in_front(self.connection_id)


class Port:
    display_name = ''
    group_id = -1
    portgroup_id = 0
    prevent_stereo = False
    last_digit_to_add = ''
    in_canvas = False
    order = None
    uuid = 0 # will contains the real JACK uuid

    # given by JACK metadatas
    pretty_name = ''
    mdata_portgroup = ''
    mdata_signal_type = ''

    def __init__(self, manager: 'PatchbayManager', port_id: int, name: str,
                 port_type: PortType, flags: int, uuid: int):
        self.manager = manager
        self.port_id = port_id
        self.full_name = name
        self.type = port_type
        self.flags = flags
        self.uuid = uuid
        self.subtype = PortSubType.REGULAR

        if (self.type is PortType.AUDIO_JACK
                and self.flags & JackPortFlag.IS_CONTROL_VOLTAGE):
            self.subtype = PortSubType.CV
        elif (self.type is PortType.MIDI_JACK
                and self.full_name.startswith(('a2j:', 'Midi-Bridge:'))):
            self.subtype = PortSubType.A2J

    def mode(self) -> PortMode:
        if self.flags & JackPortFlag.IS_OUTPUT:
            return PortMode.OUTPUT
        elif self.flags & JackPortFlag.IS_INPUT:
            return PortMode.INPUT
        else:
            return PortMode.NULL

    def full_type(self) -> tuple[PortType, PortSubType]:
        return (self.type, self.subtype)
    
    def short_name(self) -> str:
        if (self.type is PortType.MIDI_ALSA
                and self.full_name.startswith((':ALSA_IN:', ':ALSA_OUT:'))):
            return ':'.join(self.full_name.split(':')[5:])
        
        if self.full_name.startswith('a2j:'):
            long_name = self.full_name.partition(':')[2]
            if ': ' in long_name:
                # normal case for a2j
                return long_name.partition(': ')[2]

        if self.full_name.startswith('Midi-Bridge:'):
            # suppress 'Midi-Bridge:' at port name begginning
            long_name = self.full_name.partition(':')[2]
            if ') ' in long_name:
                # normal case, name is after ') '
                return long_name.partition(') ')[2]

            if ': ' in long_name:
                # pipewire jack.filter_name = True
                # Midi-bridge names starts with 'MidiBridge:ClientName:'
                return long_name.partition(': ')[2]

        return self.full_name.partition(':')[2]

    def add_the_last_digit(self):
        self.display_name += ' ' + self.last_digit_to_add
        self.last_digit_to_add = ''
        self.rename_in_canvas()

    def add_to_canvas(self):
        if self.manager.very_fast_operation:
            return
        
        if self.in_canvas:
            return

        if not self.manager.port_type_shown(self.full_type()):
            return

        display_name = self.display_name
        
        if self.pretty_name:
            display_name = self.pretty_name
        
        if not self.manager.use_graceful_names:
            display_name = self.short_name()

        self.in_canvas = True

        patchcanvas.add_port(
            self.group_id, self.port_id, display_name,
            self.mode(), self.type, self.subtype)

    def remove_from_canvas(self):
        if self.manager.very_fast_operation:
            return
        
        if not self.in_canvas:
            return

        patchcanvas.remove_port(self.group_id, self.port_id)
        self.in_canvas = False

    def rename_in_canvas(self):
        if not self.in_canvas:
            return

        display_name = self.display_name
        if self.pretty_name:
            display_name = self.pretty_name

        if not self.manager.use_graceful_names:
            display_name = self.short_name()

        patchcanvas.rename_port(
            self.group_id, self.port_id, display_name)

    def select_in_canvas(self):
        if not self.in_canvas:
            return
        
        patchcanvas.select_port(self.group_id, self.port_id)

    def __lt__(self, other: 'Port'):
        if self.type != other.type:
            return (self.type < other.type)

        if self.subtype is not other.subtype:
            return self.subtype < other.subtype

        # if self.mode() != other.mode():
        #     return (self.mode() < other.mode())

        if self.order is None and other.order is None:
            return self.port_id < other.port_id
        if self.order is None:
            return False
        if other.order is None:
            return True

        return bool(self.order < other.order)


class Portgroup:
    # Portgroup is a stereo pair of ports
    # but could be a group of more ports
    def __init__(self, manager: 'PatchbayManager', group_id: int,
                 portgroup_id: int, port_mode: PortMode, ports: tuple[Port]):
        self.manager = manager
        self.group_id = group_id
        self.portgroup_id = portgroup_id
        self.port_mode = port_mode
        self.ports = tuple(ports)

        self.mdata_portgroup = ''
        self.above_metadatas = False

        self.in_canvas = False

        for port in self.ports:
            port.portgroup_id = portgroup_id

    def port_type(self):
        if not self.ports:
            return PortType.NULL

        return self.ports[0].type

    def full_type(self) -> tuple[PortType, PortSubType]:
        if not self.ports:
            return (PortType.NULL, PortSubType.REGULAR)
        
        return self.ports[0].full_type()

    def update_ports_in_canvas(self):
        for port in self.ports:
            port.rename_in_canvas()

    def sort_ports(self):
        port_list = list(self.ports)
        port_list.sort()
        self.ports = tuple(port_list)

    def add_to_canvas(self):
        if self.manager.very_fast_operation:
            return

        if self.in_canvas:
            return
    
        if not self.manager.port_type_shown(self.full_type()):
            return

        if len(self.ports) < 2:
            return

        for port in self.ports:
            if not port.in_canvas:
                return

        self.in_canvas = True

        patchcanvas.add_portgroup(
            self.group_id, self.portgroup_id,
            self.port_mode, self.ports[0].type, self.ports[0].subtype,
            [port.port_id for port in self.ports])

    def remove_from_canvas(self):
        if self.manager.very_fast_operation:
            return
        
        if not self.in_canvas:
            return

        patchcanvas.remove_portgroup(self.group_id, self.portgroup_id)
        self.in_canvas = False


class Group:
    def __init__(self, manager: 'PatchbayManager', group_id: int,
                 name: str, group_position: GroupPos):
        self.manager = manager
        self.group_id = group_id
        self.name = name
        self.display_name = name
        self.ports = list[Port]()
        self.portgroups = list[Portgroup]()
        self._is_hardware = False
        self.client_icon = ''
        self.a2j_group = False
        self.in_canvas = False
        self.current_position = group_position
        self.uuid = 0
        
        self.has_gui = False
        self.gui_visible = False

        self.mdata_icon = ''

    def __repr__(self) -> str:
        return f"Group({self.name})"

    def update_ports_in_canvas(self):
        for port in self.ports:
            port.rename_in_canvas()

    def add_to_canvas(self, split=BoxSplitMode.UNDEF):
        if self.in_canvas:
            return

        box_type, icon_name = self._get_box_type_and_icon()

        do_split = bool(self.current_position.flags & GroupPosFlag.SPLITTED)
        split = BoxSplitMode.YES if do_split else BoxSplitMode.NO

        gpos = self.current_position

        self.display_name = self.display_name.replace('.0/', '/').replace('_', ' ')
        
        display_name = self.name
        if self.manager.use_graceful_names:
            display_name = self.display_name
        
        layout_modes_ = dict[PortMode, BoxLayoutMode]()
        for port_mode in (PortMode.INPUT, PortMode.OUTPUT, PortMode.BOTH):
            layout_modes_[port_mode] = gpos.get_layout_mode(port_mode.value)
        
        patchcanvas.add_group(
            self.group_id, display_name, split,
            box_type, icon_name, layout_modes=layout_modes_,
            null_xy=gpos.null_xy, in_xy=gpos.in_xy, out_xy=gpos.out_xy)

        self.in_canvas = True

        if do_split:
            gpos.flags |= GroupPosFlag.HAS_BEEN_SPLITTED
            patchcanvas.wrap_group_box(
                self.group_id, PortMode.INPUT,
                bool(gpos.flags & GroupPosFlag.WRAPPED_INPUT),
                animate=False)
            patchcanvas.wrap_group_box(
                self.group_id, PortMode.OUTPUT,
                bool(gpos.flags & GroupPosFlag.WRAPPED_OUTPUT),
                animate=False)
        else:
            patchcanvas.wrap_group_box(
                self.group_id, PortMode.NULL,
                bool(gpos.flags & GroupPosFlag.WRAPPED_INPUT
                     and gpos.flags & GroupPosFlag.WRAPPED_OUTPUT),
                animate=False)
            
        if self.has_gui:
            patchcanvas.set_optional_gui_state(self.group_id, self.gui_visible)

    def remove_from_canvas(self):
        if not self.in_canvas:
            return

        patchcanvas.remove_group(self.group_id)
        self.in_canvas = False

    def redraw_in_canvas(self):
        if not self.in_canvas:
            return
        
        patchcanvas.redraw_group(self.group_id)

    def update_name_in_canvas(self):
        if not self.in_canvas:
            return
        
        display_name = self.name
        if self.manager.use_graceful_names:
            display_name = self.display_name
        
        patchcanvas.rename_group(self.group_id, display_name)

    def _get_box_type_and_icon(self) -> tuple[BoxType, str]:
        box_type = BoxType.APPLICATION
        icon_name = self.name.partition('.')[0].lower()

        if self._is_hardware:
            box_type = BoxType.HARDWARE
            icon_name = ''
            if self.a2j_group or self.display_name in ("Midi-Bridge", "a2j"):
                icon_name = "a2j"

        if self.client_icon:
            box_type = BoxType.CLIENT
            icon_name = self.client_icon

        if (self.name.startswith("PulseAudio ")
                and not self.client_icon):
            if "sink" in self.name.lower():
                box_type = BoxType.MONITOR
                icon_name = 'monitor_playback'
            elif "source" in self.name.lower():
                box_type = BoxType.MONITOR
                icon_name = 'monitor_capture'

        elif (self.name.endswith(" Monitor")
                and not self.client_icon):
            # this group is (probably) a pipewire Monitor group
            box_type = BoxType.MONITOR
            icon_name = 'monitor_playback'
        
        if self.mdata_icon:
            icon_name = self.mdata_icon

        return (box_type, icon_name)

    def semi_hide(self, yesno: bool):
        if not self.in_canvas:
            return 
        
        patchcanvas.semi_hide_group(self.group_id, yesno)

    def set_in_front(self):
        if not self.in_canvas:
            return
        
        patchcanvas.set_group_in_front(self.group_id)

    def get_number_of_boxes(self) -> int:
        if not self.in_canvas:
            return 0

        return patchcanvas.get_number_of_boxes(self.group_id)

    def select_filtered_box(self, n_select=0):
        if not self.in_canvas:
            return
        
        patchcanvas.select_filtered_group_box(self.group_id, n_select)

    def set_optional_gui_state(self, visible: bool):
        self.has_gui = True
        self.gui_visible = visible
        
        if not self.in_canvas:
            return
        
        patchcanvas.set_optional_gui_state(self.group_id, visible)

    def remove_all_ports(self):
        if self.in_canvas:
            for portgroup in self.portgroups:
                portgroup.remove_from_canvas()

            for port in self.ports:
                port.remove_from_canvas()

        self.portgroups.clear()
        self.ports.clear()

    def add_port(self, port: Port):
        port.group_id = self.group_id
        port_full_name = port.full_name

        if (port_full_name.startswith('a2j:')
                and not port.flags & JackPortFlag.IS_PHYSICAL):
            port_full_name = port_full_name.partition(':')[2]
            
        elif port.type is PortType.MIDI_ALSA:
            port_full_name = ':'.join(port_full_name.split(':')[4:])

        port.display_name = port_full_name.partition(':')[2]

        if not self.ports:
            # we are adding the first port of the group
            if port.flags & JackPortFlag.IS_PHYSICAL:
                self._is_hardware = True

            if not self.current_position.fully_set:
                if self._is_hardware:
                    self.current_position.flags |= GroupPosFlag.SPLITTED
                self.current_position.fully_set = True
                self.save_current_position()

        self.ports.append(port)
        self.manager._ports_by_name[port.full_name] = port

    def remove_port(self, port: Port):
        if port in self.ports:
            port.remove_from_canvas()
            self.ports.remove(port)
        
        if self.manager._ports_by_name.get(port.full_name):
            self.manager._ports_by_name.pop(port.full_name)

    def remove_portgroup(self, portgroup: Portgroup):
        if portgroup in self.portgroups:
            portgroup.remove_from_canvas()
            for port in portgroup.ports:
                port.portgroup_id = 0
            self.portgroups.remove(portgroup)

    def portgroup_memory_added(self, portgroup_mem: PortgroupMem):
        if portgroup_mem.group_name != self.name:
            return

        remove_set = set[Portgroup]()

        # first remove any existing portgroup with one of the porgroup_mem ports
        for portgroup in self.portgroups:
            if (portgroup.port_mode is not portgroup_mem.port_mode
                    or portgroup.port_type() is not portgroup_mem.port_type):
                continue

            for port in portgroup.ports:
                if port.short_name() in portgroup_mem.port_names:
                    remove_set.add(portgroup)

        for portgroup in remove_set:
            self.remove_portgroup(portgroup)

        # add a portgroup if all needed ports are present and consecutive
        port_list = list[Port]()

        for port in self.ports:
            if (port.mode() is not portgroup_mem.port_mode
                    or port.type is not portgroup_mem.port_type):
                continue

            if port.short_name() == portgroup_mem.port_names[len(port_list)]:
                port_list.append(port)

                if len(port_list) == len(portgroup_mem.port_names):
                    # all ports are presents, create the portgroup
                    portgroup = self.manager.new_portgroup(
                        self.group_id, port.mode(), port_list)
                    self.portgroups.append(portgroup)
                    portgroup.add_to_canvas()
                    break

            elif port_list:
                # here it is a port breaking the consecutivity of the portgroup
                break

    def save_current_position(self):
        self.manager.save_group_position(self.current_position)

    def set_group_position(self, group_position: GroupPos, view_change=False):
        if not self.in_canvas:
            return

        ex_gpos_flags = self.current_position.flags
        self.current_position = group_position
        gpos = self.current_position

        for port_mode, layout_mode in group_position.layout_modes.items():
            patchcanvas.set_group_layout_mode(
                self.group_id, port_mode, layout_mode)

        patchcanvas.move_group_boxes(
            self.group_id, gpos.null_xy, gpos.in_xy, gpos.out_xy)

        prevent_overlap = not view_change

        # restore split and wrapped modes
        if gpos.flags & GroupPosFlag.SPLITTED:
            if not ex_gpos_flags & GroupPosFlag.SPLITTED:
                patchcanvas.split_group(self.group_id)

            patchcanvas.wrap_group_box(
                self.group_id, PortMode.INPUT,
                bool(gpos.flags & GroupPosFlag.WRAPPED_INPUT),
                prevent_overlap=prevent_overlap)
            patchcanvas.wrap_group_box(
                self.group_id, PortMode.OUTPUT,
                bool(gpos.flags & GroupPosFlag.WRAPPED_OUTPUT),
                prevent_overlap=prevent_overlap)
        else:
            patchcanvas.wrap_group_box(
                self.group_id, PortMode.NULL,
                bool(gpos.flags & (GroupPosFlag.WRAPPED_INPUT
                                   | GroupPosFlag.WRAPPED_OUTPUT)),
                prevent_overlap=prevent_overlap)

            if ex_gpos_flags & GroupPosFlag.SPLITTED:
                patchcanvas.animate_before_join(self.group_id)

    def set_layout_mode(self, port_mode: PortMode, layout_mode: BoxLayoutMode):
        self.current_position.set_layout_mode(port_mode, layout_mode)
        self.save_current_position()

        if not self.in_canvas:
            return
        
        patchcanvas.set_group_layout_mode(self.group_id, port_mode, layout_mode)

    def wrap_box(self, port_mode: int, yesno: bool):
        wrap_flag = GroupPosFlag.WRAPPED_OUTPUT | GroupPosFlag.WRAPPED_INPUT
        if port_mode == PortMode.INPUT:
            wrap_flag = GroupPosFlag.WRAPPED_INPUT
        elif port_mode == PortMode.OUTPUT:
            wrap_flag = GroupPosFlag.WRAPPED_OUTPUT

        if yesno:
            self.current_position.flags |= wrap_flag
        else:
            self.current_position.flags &= ~wrap_flag

        self.save_current_position()

        if not self.in_canvas:
            return

        patchcanvas.wrap_group_box(self.group_id, port_mode, yesno)

    def set_client_icon(self, icon_name: str, from_metadata=False):
        if from_metadata:
            self.mdata_icon = icon_name
        else:
            self.client_icon = icon_name
        
        box_type, ex_icon_name = self._get_box_type_and_icon()
        
        if self.in_canvas:
            patchcanvas.set_group_icon(
                self.group_id, box_type, icon_name)

    def get_pretty_client(self) -> str:
        for client_name in ('firewire_pcm', 'a2j',
                            'Hydrogen', 'ardour', 'Ardour', 'Mixbus', 'mixbus',
                            'Qtractor', 'SooperLooper', 'sooperlooper', 'Luppp',
                            'seq64', 'calfjackhost', 'rakarrack-plus',
                            'seq192', 'Non-Mixer', 'Non-Mixer-XT', 'jack_mixer'):
            if self.name == client_name:
                return client_name

            if self.name.startswith(client_name + '.'):
                return client_name
            
            name = self.name.partition('/')[0]
            if name == client_name:
                return client_name
            
            if name.startswith(client_name + '_'):
                if name.replace(client_name + '_', '', 1).isdigit():
                    return client_name
            
            if ' (' in name and name.endswith(')'):
                name = name.partition(' (')[0]
                if name == client_name:
                    return client_name
                
                if name.startswith(client_name + '_'):
                    if name.replace(client_name + '_', '', 1).isdigit():
                        return client_name

        return ''

    def graceful_port(self, port: Port):
        def split_end_digits(name: str) -> tuple[str, str]:
            num = ''
            while name and name[-1].isdigit():
                num = name[-1] + num
                name = name[:-1]

            if num.startswith('0') and num not in ('0', '09'):
                num = num[1:]

            return (name, num)

        def cut_end(name: str, *ends: str) -> str:
            for end in ends:
                if name.endswith(end):
                    return name.rsplit(end)[0]
            return name

        client_name = self.get_pretty_client()

        if (not client_name
                and ((port.type is PortType.MIDI_JACK
                        and port.full_name.startswith(
                            ('a2j:', 'Midi-Bridge:')))
                     or port.type is PortType.MIDI_ALSA)
                and port.flags & JackPortFlag.IS_PHYSICAL):
            client_name = 'a2j'

        display_name = port.short_name()
        s_display_name = display_name

        if client_name == 'firewire_pcm':
            if '(' in display_name and ')' in display_name:
                after_para = display_name.partition('(')[2]
                display_name = after_para.rpartition(')')[0]
                display_name, num = split_end_digits(display_name)

                if num:
                    if display_name.endswith(':'):
                        display_name = display_name[:-1]
                    display_name += ' ' + num
            else:
                display_name = display_name.partition('_')[2]
                display_name = cut_end(display_name, '_in', '_out')
                display_name = display_name.replace(':', ' ')
                display_name, num = split_end_digits(display_name)
                display_name = display_name + num

        elif client_name == 'Hydrogen':
            if display_name.startswith('Track_'):
                display_name = display_name.replace('Track_', '', 1)

                num, udsc, name = display_name.partition('_')
                if num.isdigit():
                    display_name = num + ' ' + name

            if display_name.endswith('_Main_L'):
                display_name = display_name.replace('_Main_L', ' L', 1)
            elif display_name.endswith('_Main_R'):
                display_name = display_name.replace('_Main_R', ' R', 1)

        elif client_name == 'a2j':
            display_name, num = split_end_digits(display_name)
            if num:
                if display_name.endswith(' MIDI '):
                    display_name = cut_end(display_name, ' MIDI ')

                    if num == '1':
                        port.last_digit_to_add = '1'
                    else:
                        display_name += ' ' + num

                elif display_name.endswith(' Port-'):
                    display_name = cut_end(display_name, ' Port-')

                    if num == '0':
                        port.last_digit_to_add = '0'
                    else:
                        display_name += ' ' + num

        elif client_name in ('ardour', 'Ardour', 'Mixbus', 'mixbus'):
            if '/TriggerBox/' in display_name:
                display_name = '▸ ' + display_name.replace('/TriggerBox/', '/', 1)
            
            for pt in ('audio', 'midi'):
                if display_name == f"physical_{pt}_input_monitor_enable":
                    display_name = "physical monitor"
                    break
            else:
                display_name, num = split_end_digits(display_name)
                if num:
                    display_name = cut_end(display_name,
                                        '/audio_out ', '/audio_in ',
                                        '/midi_out ', '/midi_in ')
                    if num == '1':
                        port.last_digit_to_add = '1'
                    else:
                        display_name += ' ' + num

        elif client_name == 'Qtractor':
            display_name, num = split_end_digits(display_name)
            if num:
                display_name = cut_end(display_name,
                                       '/in_', '/out_')
                if num == '1':
                    port.last_digit_to_add = '1'
                else:
                    display_name += ' ' + num
        
        elif client_name in ('Non-Mixer', 'Non-Mixer-XT'):
            display_name, num = split_end_digits(display_name)
            if num:
                display_name = cut_end(display_name, '/in-', '/out-')
                
                if num == '1':
                    port.last_digit_to_add = '1'
                else:
                    display_name += ' ' + num
        
        elif client_name == 'jack_mixer':
            prefix, out, side = display_name.rpartition(' Out')
            if out and side in (' L', ' R', ''):
                display_name = prefix + side
                        
        elif client_name in ('SooperLooper', 'sooperlooper'):
            display_name, num = split_end_digits(display_name)
            if num:
                display_name = cut_end(display_name,
                                       '_in_', '_out_')
                if num == '1':
                    port.last_digit_to_add = '1'
                else:
                    display_name += ' ' + num

        elif client_name == 'Luppp':
            if display_name.endswith('\n'):
                display_name = display_name[:-1]

            display_name = display_name.replace('_', ' ')

        elif client_name == 'seq64':
            display_name = display_name.replace('seq64 midi ', '', 1)

        elif client_name == 'seq192':
            display_name = display_name.replace('seq192 ', '', 1)

        elif client_name == 'calfjackhost':
            display_name, num = split_end_digits(display_name)
            if num:
                display_name = cut_end(display_name,
                                       ' Out #', ' In #')

                display_name += " " + num

        elif client_name == 'rakarrack-plus':
            if display_name.startswith(('rakarrack-plus ', 'rakarrack-plus.')):
                display_name = display_name[15:]
            display_name = display_name.replace('_', ' ')

        elif not client_name:
            display_name = display_name.replace('_', ' ')
            if display_name.lower().endswith(('-left', ' left')):
                display_name = display_name[:-5] + ' L'
            elif display_name.lower().endswith(('-right', ' right')):
                display_name = display_name[:-6] + ' R'
            elif display_name.lower() == 'left in':
                display_name = 'In L'
            elif display_name.lower() == 'right in':
                display_name = 'In R'
            elif display_name.lower() == 'left out':
                display_name = 'Out L'
            elif display_name.lower() == 'right out':
                display_name = 'Out R'

            if display_name.startswith('Audio'):
                display_name = display_name.replace('Audio ', '')

        # reduce graceful name for pipewire Midi-Bridge with
        # option jack.filter_name = true
        if (port.full_name.startswith('Midi-Bridge')
                and display_name.startswith(('capture_', 'playback_'))):
            display_name = display_name.partition('_')[2]

        port.display_name = display_name if display_name else s_display_name

    def add_portgroup(self, portgroup: Portgroup):
        self.portgroups.append(portgroup)

    def change_port_types_view(self):
        # first add group to canvas if not already
        self.add_to_canvas()

        for portgroup in self.portgroups:
            if not self.manager.port_type_shown(portgroup.full_type()):
                portgroup.remove_from_canvas()

        for port in self.ports:
            if not self.manager.port_type_shown(port.full_type()):
                port.remove_from_canvas()

        for port in self.ports:
            port.add_to_canvas()

        for portgroup in self.portgroups:
            portgroup.add_to_canvas()

        # remove group from canvas if no visible ports
        for port in self.ports:
            if port.in_canvas:
                break
        else:
            self.remove_from_canvas()

    def stereo_detection(self, port: Port) -> Union[Port, None]:
        if port.type != PortType.AUDIO_JACK or port.subtype != PortSubType.REGULAR:
            return

        # find the last port with same type and mode in the group
        for other_port in reversed(self.ports):
            if other_port == port:
                continue

            if (other_port.type == port.type
                    and other_port.subtype == port.subtype
                    and other_port.mode() == port.mode()
                    and not other_port.portgroup_id
                    and not other_port.prevent_stereo):
                for portgroup_mem in self.manager.portgroups_memory:
                    if (portgroup_mem.group_name == self.name
                        and portgroup_mem.port_mode == other_port.mode()
                        and portgroup_mem.port_type == other_port.type
                        and other_port.short_name() in portgroup_mem.port_names):
                        # other_port (left) is in a remembered portgroup
                        # prevent stereo detection
                        return
                break
        else:
            return

        may_match_set = set[str]()

        port_name = port.full_name.replace(self.name + ':', '', 1)
        other_port_name = other_port.full_name.replace(self.name + ':', '', 1)

        if not (port_name and other_port_name):
            return

        if port.flags & JackPortFlag.IS_PHYSICAL:
            # force stereo detection for system ports
            # it forces it for firewire long and strange names
            may_match_set.add(other_port_name)

        elif port_name[-1].isdigit():
            # Port ends with digit
            base_port = port_name[:-1]
            in_num = port_name[-1]

            while base_port and base_port[-1].isdigit():
                in_num = base_port[-1] + in_num
                base_port = base_port[:-1]

            # if Port ends with Ldigits or Rdigits
            if base_port.endswith('R'):
                may_match_set.add(base_port[:-1] + 'L' + in_num)
            else:
                may_match_set.add(base_port + str(int(in_num) -1))

                if int(in_num) in (1, 2):
                    if base_port.endswith((' ', ('_'))):
                        may_match_set.add(base_port[:-1])
                    else:
                        may_match_set.add(base_port)
        else:
            # Port ends with non digit
            if port_name.endswith('R'):
                may_match_set.add(port_name[:-1] + 'L')
                if len(port_name) >= 2:
                    if port_name[-2] == ' ':
                        may_match_set.add(port_name[:-2])
                    else:
                        may_match_set.add(port_name[:-1])

            elif port_name.endswith('right'):
                may_match_set.add(port_name[:-5] + 'left')

            elif port_name.endswith('Right'):
                may_match_set.add(port_name[:-5] + 'Left')

            elif port_name.endswith('(Right)'):
                may_match_set.add(port_name[:-7] + '(Left)')

            elif port_name.endswith('.r'):
                may_match_set.add(port_name[:-2] + '.l')

            elif port_name.endswith('_r'):
                may_match_set.add(port_name[:-2] + '_l')

            elif port_name.endswith('_r\n'):
                may_match_set.add(port_name[:-3] + '_l\n')

            for x in ('out', 'Out', 'output', 'Output', 'in', 'In',
                      'input', 'Input', 'audio input', 'audio output'):
                if port_name.endswith('R ' + x):
                    may_match_set.add('L ' + x)

                elif port_name.endswith('right ' + x):
                    may_match_set.add('left ' + x)

                elif port_name.endswith('Right ' + x):
                    may_match_set.add('Left ' + x)

        if other_port_name in may_match_set:
            return other_port

    def check_for_portgroup_on_last_port(self):
        if not self.ports:
            return

        last_port = self.ports[-1]
        last_port_name = last_port.short_name()

        # check in the saved portgroups if we need to make a portgroup
        # or prevent stereo detection
        for portgroup_mem in self.manager.portgroups_memory:
            if (portgroup_mem.group_name == self.name
                    and portgroup_mem.port_type == last_port.type
                    and portgroup_mem.port_mode == last_port.mode()
                    and last_port_name == portgroup_mem.port_names[-1]):
                if (len(portgroup_mem.port_names) == 1
                    or portgroup_mem.port_names.index(last_port_name) + 1
                        != len(portgroup_mem.port_names)):
                    return

                port_list = list[Port]()

                for port in self.ports:
                    if (port.type == last_port.type
                            and port.mode() == last_port.mode()):
                        if (len(portgroup_mem.port_names) > len(port_list)
                                and port.short_name()
                                == portgroup_mem.port_names[len(port_list)]):
                            port_list.append(port)

                            if len(port_list) == len(portgroup_mem.port_names):
                                portgroup = self.manager.new_portgroup(
                                    self.group_id, port.mode(), port_list)
                                self.portgroups.append(portgroup)
                                for port in port_list:
                                    if not port.in_canvas:
                                        break
                                else:
                                    portgroup.add_to_canvas()

                        elif port_list:
                            return

        # detect left audio port if it is a right one
        other_port = self.stereo_detection(last_port)
        if other_port is not None:
            portgroup = self.manager.new_portgroup(
                self.group_id, last_port.mode(), (other_port, last_port))
            self.add_portgroup(portgroup)

            if self.in_canvas:
                portgroup.add_to_canvas()

    def check_for_display_name_on_last_port(self):
        if not self.ports:
            return

        last_port = self.ports[-1]
        last_digit = last_port.full_name[-1]

        if last_digit not in ('1', '2'):
            return

        for port in reversed(self.ports[:-1]):
            if (port.type == last_port.type
                    and port.mode() == last_port.mode()
                    and port is not last_port):
                if (port.full_name[:-1] == last_port.full_name[:-1]
                        and ((port.last_digit_to_add == '0'
                              and last_digit == '1'))
                             or (port.last_digit_to_add == '1'
                                 and last_digit == '2')):
                        port.add_the_last_digit()
                break

    def sort_ports_in_canvas(self):
        already_optimized = self.manager.optimized_operation
        self.manager.optimize_operation(True)

        conn_list = list[Connection]()

        if not self.manager.very_fast_operation:
            for conn in self.manager.connections:
                for port in self.ports:
                    if (port in (conn.port_out, conn.port_in)
                            and conn not in conn_list):
                        conn_list.append(conn)
            
            for connection in conn_list:
                connection.remove_from_canvas()
            
            for portgroup in self.portgroups:
                portgroup.remove_from_canvas()

            for port in self.ports:
                port.remove_from_canvas()
        
        self.ports.sort()

        # search and remove existing portgroups with non consecutive ports
        portgroups_to_remove = list[Portgroup]()

        for portgroup in self.portgroups:
            search_index = 0
            previous_port = None
            seems_ok = False

            for port in self.ports:
                if not seems_ok and port is portgroup.ports[search_index]:
                    if (port.mdata_portgroup != portgroup.mdata_portgroup
                            and not portgroup.above_metadatas):
                        portgroups_to_remove.append(portgroup)
                        break

                    if (not portgroup.above_metadatas and not search_index
                            and previous_port is not None
                            and previous_port.mdata_portgroup
                            and previous_port.mdata_portgroup == port.mdata_portgroup):
                        # previous port had the same portgroup metadata
                        # that this port. we need to remove this portgroup.
                        portgroups_to_remove.append(portgroup)
                        break

                    search_index += 1
                    if search_index == len(portgroup.ports):
                        # all ports of portgroup are consecutive
                        # but still exists the risk that metadatas says
                        # that the portgroup has now more ports
                        seems_ok = True
                        if (portgroup.above_metadatas
                                or not portgroup.mdata_portgroup):
                            break

                elif search_index:
                    if (seems_ok
                            and (port.mdata_portgroup != previous_port.mdata_portgroup
                                 or port.type != portgroup.port_type()
                                 or port.mode() != portgroup.port_mode)):
                        # port after the portgroup has not to make
                        # the portgroup higher. We keep this portgroup
                        break

                    # this port breaks portgroup ports consecutivity.
                    # note that ports have been just sorted by type and mode
                    # so no risk that this port is falsely breaking portgroup
                    portgroups_to_remove.append(portgroup)
                    break

                previous_port = port
            else:
                if not seems_ok:
                    portgroups_to_remove.append(portgroup)

        for portgroup in portgroups_to_remove:
            self.remove_portgroup(portgroup)

        # add missing portgroups aboving metadatas from portgroup memory
        for portgroup_mem in self.manager.portgroups_memory:
            if not portgroup_mem.above_metadatas:
                continue

            if portgroup_mem.group_name != self.name:
                continue

            founded_ports = list[Port]()

            for port in self.ports:
                if (not port.portgroup_id
                        and port.type == portgroup_mem.port_type
                        and port.mode() == portgroup_mem.port_mode
                        and port.short_name()
                            == portgroup_mem.port_names[len(founded_ports)]):
                    founded_ports.append(port)
                    if len(founded_ports) == len(portgroup_mem.port_names):
                        new_portgroup = self.manager.new_portgroup(
                            self.group_id, port.mode(), founded_ports)
                        self.portgroups.append(new_portgroup)
                        break

                elif founded_ports:
                    break

        # detect and add portgroups given from metadatas
        portgroups_mdata = list[dict]() # list of dicts

        for port in self.ports:
            if port.mdata_portgroup:
                pg_mdata = None
                if portgroups_mdata:
                    pg_mdata = portgroups_mdata[-1]

                if not port.portgroup_id:
                    if (pg_mdata is not None
                            and pg_mdata['pg_name'] == port.mdata_portgroup
                            and pg_mdata['port_type'] == port.type
                            and pg_mdata['port_mode'] == port.mode()):
                        pg_mdata['ports'].append(port)
                    else:
                        portgroups_mdata.append(
                            {'pg_name': port.mdata_portgroup,
                             'port_type': port.type,
                             'port_mode': port.mode(),
                             'ports':[port]})
        
        for pg_mdata in portgroups_mdata:
            if len(pg_mdata['ports']) < 2:
                continue

            new_portgroup = self.manager.new_portgroup(
                self.group_id, pg_mdata['port_mode'], pg_mdata['ports'])
            new_portgroup.mdata_portgroup = pg_mdata['pg_name']
            self.portgroups.append(new_portgroup)
        
        # add missing portgroups from portgroup memory
        for portgroup_mem in self.manager.portgroups_memory:
            if portgroup_mem.above_metadatas:
                continue

            if portgroup_mem.group_name != self.name:
                continue

            founded_ports = list[Port]()

            for port in self.ports:
                if (not port.portgroup_id
                        and port.type == portgroup_mem.port_type
                        and port.mode() == portgroup_mem.port_mode
                        and port.short_name()
                            == portgroup_mem.port_names[len(founded_ports)]):
                    founded_ports.append(port)
                    if len(founded_ports) == len(portgroup_mem.port_names):
                        new_portgroup = self.manager.new_portgroup(
                            self.group_id, port.mode(), founded_ports)
                        self.portgroups.append(new_portgroup)
                        break

                elif founded_ports:
                    break
        
        if not self.manager.very_fast_operation:
            # ok for re-adding all items to canvas
            for port in self.ports:
                port.add_to_canvas()

            for portgroup in self.portgroups:
                portgroup.add_to_canvas()
        
            for connection in conn_list:
                connection.add_to_canvas()
        
        if not already_optimized:
            self.manager.optimize_operation(False)
            self.redraw_in_canvas()

    def add_all_ports_to_canvas(self):
        for port in self.ports:
            port.add_to_canvas()

        for portgroup in self.portgroups:
            portgroup.add_to_canvas()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# PatchBay Canvas engine using QGraphicsView/Scene
# Copyright (C) 2010-2019 Filipe Coelho <falktx@falktx.com>
# Copyright (C) 2019-2022 Mathieu Picot <picotmathieu@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.

import logging
import os

from PyQt5.QtCore import QRectF, QFile
from PyQt5.QtGui import QPainter, QIcon
from PyQt5.QtSvg import QGraphicsSvgItem, QSvgRenderer
from PyQt5.QtWidgets import QGraphicsPixmapItem

from .init_values import canvas, CanvasItemType, BoxType, PortMode


_logger = logging.getLogger(__name__)
_app_icons_cache = {}


def get_app_icon(icon_name: str) -> QIcon:
    if icon_name in _app_icons_cache.keys():
        return _app_icons_cache[icon_name]
    
    icon = QIcon.fromTheme(icon_name)

    if icon.isNull():
        for ext in ('svg', 'svgz', 'png'):
            filename = ":app_icons/%s.%s" % (icon_name, ext)

            if QFile.exists(filename):
                del icon
                icon = QIcon()
                icon.addFile(filename)
                break

    if icon.isNull():
        for path in ('/usr/local', '/usr', '%s/.local' % os.getenv('HOME')):
            for ext in ('png', 'svg', 'svgz', 'xpm'):
                filename = "%s/share/pixmaps/%s.%s" % (path, icon_name, ext)
                if QFile.exists(filename):
                    del icon
                    icon = QIcon()
                    icon.addFile(filename)
                    break

    _app_icons_cache[icon_name] = icon

    return icon


class IconPixmapWidget(QGraphicsPixmapItem):
    def __init__(self, box_type: BoxType, icon_name: str, parent):
        QGraphicsPixmapItem.__init__(self, parent)

        self._size = QRectF(0.0, 0.0, 24.0, 24.0)
        self.icon = None
        self.x_offset = 4
        self.y_offset = 4

        self.set_icon(box_type, icon_name)

    def set_icon(self, box_type: BoxType, name: str, port_mode=PortMode.NULL):
        self.icon = get_app_icon(name)

        if not self.icon.isNull():
            scale = canvas.scene.get_zoom_scale()
            pixmap = self.icon.pixmap(int(0.5 + 24 * scale),
                                      int(0.5 + 24 * scale))
            self.setPixmap(pixmap)
            self.setScale(1.0 / scale)
            self.setPos(4.0, 4.0)

    def update_zoom(self, scale: float):
        if self.icon is None or scale <= 0.0:
            return

        pixmap = self.icon.pixmap(int(0.5 + 24 * scale), int(0.5 + 24 * scale))
        self.setPixmap(pixmap)
        self.setScale(1.0 / scale)

    def is_null(self)->bool:
        if self.icon is None:
            return True

        return self.icon.isNull()

    def set_pos(self, x: int, y: int):
        self.x_offset = x
        self.y_offset = y
        self.setPos(float(x), float(y))
        
    def type(self) -> CanvasItemType:
        return CanvasItemType.ICON


class IconSvgWidget(QGraphicsSvgItem):
    def __init__(self, box_type: BoxType, name: str, port_mode: PortMode, parent):
        QGraphicsSvgItem.__init__(self, parent)
        self._renderer = None
        self._size = QRectF(4, 4, 24, 24)
        self.set_icon(box_type, name, port_mode)

    def set_icon(self, box_type: BoxType, name: str, port_mode: PortMode):
        name = name.lower()
        icon_path = ""
        theme = canvas.theme.icon

        if box_type is BoxType.APPLICATION:
            self._size = QRectF(3, 2, 19, 18)

            if "audacious" in name:
                icon_path = ":/scalable/pb_audacious.svg"
                self._size = QRectF(5, 4, 16, 16)
            elif "clementine" in name:
                icon_path = ":/scalable/pb_clementine.svg"
                self._size = QRectF(5, 4, 16, 16)
            elif "distrho" in name:
                icon_path = ":/scalable/pb_distrho.svg"
                self._size = QRectF(5, 4, 16, 16)
            elif "jamin" in name:
                icon_path = ":/scalable/pb_jamin.svg"
                self._size = QRectF(5, 3, 16, 16)
            elif "mplayer" in name:
                icon_path = ":/scalable/pb_mplayer.svg"
                self._size = QRectF(5, 4, 16, 16)
            elif "vlc" in name:
                icon_path = ":/scalable/pb_vlc.svg"
                self._size = QRectF(5, 3, 16, 16)
            else:
                icon_path = ":/scalable/pb_generic.svg"
                self._size = QRectF(4, 4, 24, 24)

        elif box_type is BoxType.HARDWARE:
            if name == "a2j":
                icon_path = theme.hardware_midi
                self._size = QRectF(4, 4, 24, 24)
            else:
                if port_mode is PortMode.INPUT:
                    icon_path = theme.hardware_playback
                elif port_mode is PortMode.OUTPUT:
                    icon_path = theme.hardware_capture
                else:
                    icon_path = theme.hardware_grouped
                self._size = QRectF(4, 4, 24, 24)

        elif box_type is BoxType.DISTRHO:
            icon_path = ":/scalable/pb_distrho.svg"
            self._size = QRectF(5, 4, 16, 16)

        elif box_type is BoxType.FILE:
            icon_path = ":/scalable/pb_file.svg"
            self._size = QRectF(5, 4, 16, 16)

        elif box_type is BoxType.PLUGIN:
            icon_path = ":/scalable/pb_plugin.svg"
            self._size = QRectF(5, 4, 16, 16)

        elif box_type is BoxType.LADISH_ROOM:
            # TODO - make a unique ladish-room icon
            icon_path = ":/scalable/pb_hardware.svg"
            self._size = QRectF(5, 2, 16, 16)

        elif box_type is BoxType.MONITOR:
            if name == 'monitor_capture':
                icon_path = theme.monitor_capture
            elif name == 'monitor_playback':
                icon_path = theme.monitor_playback
            else:
                icon_path = ":/canvas/dark/" + name
            self._size = QRectF(4, 4, 24, 24)

        else:
            self._size = QRectF(0, 0, 0, 0)
            _logger.critical(f"set_icon({str(box_type)}, {name})"
                             " - unsupported icon requested")
            return

        self._renderer = QSvgRenderer(icon_path, canvas.scene)
        self.setSharedRenderer(self._renderer)
        self.update()
        
    def update_zoom(self, scale: float):
        pass

    def type(self) -> CanvasItemType:
        return CanvasItemType.ICON

    def is_null(self)->bool:
        return False

    def set_pos(self, x: int, y: int):
        self._size = QRectF(x, y, 24, 24)

    def boundingRect(self):
        return self._size

    def paint(self, painter, option, widget):
        if not self._renderer:
            QGraphicsSvgItem.paint(self, painter, option, widget)
            return

        painter.save()
        painter.setRenderHint(QPainter.Antialiasing, False)
        painter.setRenderHint(QPainter.TextAntialiasing, False)
        self._renderer.render(painter, self._size)
        painter.restore()

# ------------------------------------------------------------------------------------------------------------






#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# PatchBay Canvas engine using QGraphicsView/Scene
# Copyright (C) 2010-2019 Filipe Coelho <falktx@falktx.com>
# Copyright (C) 2019-2022 Mathieu Picot <picotmathieu@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the doc/GPL.txt file.


# Imports (Global)
import logging
from math import floor
from re import A
from typing import TYPE_CHECKING

from PyQt5.QtCore import Qt, QPointF, QRectF
from PyQt5.QtGui import (
    QBrush, QFontMetrics, QPainter, QPen, QPolygonF,
    QLinearGradient, QIcon, QCursor, QColor)
from PyQt5.QtWidgets import QGraphicsItem, QMenu, QApplication


# Imports (Custom)
from .init_values import (
    CanvasItemType,
    PortObject,
    PortSubType,
    canvas,
    features,
    CallbackAct,
    PortMode,
    PortType)

from .utils import canvas_callback
from .connectable_widget import ConnectableWidget
from .connect_menu import ConnectMenu, ConnectableContextMenu
from .line_widget import LineWidget

if TYPE_CHECKING:
    from .box_widget import BoxWidget
    from .portgroup_widget import PortgroupWidget
    

# --------------------
_translate = QApplication.translate

# --------------------

class PortWidget(ConnectableWidget):
    def __init__(self, port: PortObject, parent: 'BoxWidget'):
        ConnectableWidget.__init__(self, port, parent)        
        self._logger = logging.getLogger(__name__)

        # Save Variables, useful for later
        self._port = port
        self._port_id = port.port_id
        self._port_name = port.port_name
        self._portgrp_id = port.portgrp_id
        self._pg_pos = port.pg_pos
        self._pg_len = port.pg_len
        self._port_subtype = port.port_subtype
        self._print_name = port.port_name
        self._print_name_right = ''
        self._name_truncked = False
        self._trunck_sep = '⠿'

        # Base Variables
        self._port_width = 15
        self._port_height = canvas.theme.port_height

        theme = canvas.theme.port
        if self._port_type == PortType.AUDIO_JACK:
            if self._port_subtype is PortSubType.CV:
                theme = theme.cv
            else:
                theme = theme.audio
        elif self._port_type == PortType.MIDI_JACK:
            theme = theme.midi
        
        self._theme = theme
        self._port_font = theme.font()

        self._portgrp_widget = None
        self._loop_select_done = False

        self._lines_widgets = list[LineWidget]()
        self._connect_pos = QPointF(0.0, 0.0)
        self._update_connect_pos()

    def get_port_id(self) -> int:
        return self._port_id

    def get_connection_distance(self) -> float:
        return self._port_width

    def get_port_width(self) -> float:
        return self._port_width

    def set_portgroup_id(self, portgrp_id: int, index: int, portgrp_len: int):
        self._portgrp_id = portgrp_id
        self._pg_pos = index
        self._pg_len = portgrp_len
        self._update_connect_pos()

    def set_portgroup_widget(self, widget: 'PortgroupWidget'):
        self._portgrp_widget = widget

    def set_port_name(self, port_name: str):
        self._port_name = port_name
        self._update_connect_pos()

    def set_port_width(self, port_width):
        self._port_width = port_width
        self._update_connect_pos()

    def set_print_name(self, print_name:str, width_limited: int):
        self._print_name = print_name
        self._name_truncked = False

        if width_limited:
            #sizer = QFontMetrics(self._port_font)
            long_size = self._theme.get_text_width(self._print_name)
            
            if long_size > width_limited:
                name_len = len(self._print_name)
                middle = int(name_len / 2)
                left_text = self._print_name[:middle]
                middle_text = self._trunck_sep
                right_text = self._print_name[middle + 1:]
                left_size = self._theme.get_text_width(left_text)
                middle_size = self._theme.get_text_width(middle_text)
                right_size = self._theme.get_text_width(right_text)

                while left_size + middle_size + right_size > width_limited:
                    if left_size > right_size:
                        left_text = left_text[:-1]
                        left_size = self._theme.get_text_width(left_text)
                    else:
                        right_text = right_text[1:]
                        right_size = self._theme.get_text_width(right_text)
                        
                    if not (left_text or right_text):
                        break

                self._print_name = left_text
                self._print_name_right = right_text
                self._name_truncked = True
        self._update_connect_pos()

    def get_text_width(self):
        if self._name_truncked:
            return (self._theme.get_text_width(self._print_name)
                    + self._theme.get_text_width(self._trunck_sep)
                    + self._theme.get_text_width(self._print_name_right))
        
        return self._theme.get_text_width(self._print_name)

    def set_as_stereo(self, port_id: int):
        canvas_callback(
            CallbackAct.PORTGROUP_ADD,
            self._group_id, self._port_mode, self._port_type,
            tuple([p.port_id for p in canvas.list_ports(group_id=self._group_id)
                   if p.port_id in (self._port_id, port_id)]))

    def type(self) -> CanvasItemType:
        return CanvasItemType.PORT

    def _update_connect_pos(self):
        phi = 0.75 if self._pg_len > 2 else 0.62
        
        x_delta = 0
        if self._port_mode is PortMode.OUTPUT:
            x_delta = self._port_width + 12
        
        height = canvas.theme.port_height
        y_delta = canvas.theme.port_height / 2
        
        if self._pg_len >= 2:
            first_old_y = height * phi
            last_old_y = height * (self._pg_len - phi)
            delta = (last_old_y - first_old_y) / (self._pg_len -1)
            y_delta = (first_old_y
                       + (self._pg_pos * delta)
                       - (height * self._pg_pos))
            
        if not self.isVisible():
            # item is hidden port when its box is folded
            y_delta = height - y_delta

        self._connect_pos = QPointF(x_delta, y_delta)

    def connect_pos(self) -> QPointF:
        return self.scenePos() + self._connect_pos

    def add_line_to_port(self, line: 'LineWidget'):
        self._lines_widgets.append(line)

    def remove_line_from_port(self, line: 'LineWidget'):
        if line in self._lines_widgets:
            self._lines_widgets.remove(line)

    def setVisible(self, visible: bool):
        super().setVisible(visible)
        self._update_connect_pos()

    def itemChange(self, change: int, value: bool):
        if change == QGraphicsItem.ItemSelectedHasChanged:
            if self.changing_select_state:
                self.changing_select_state = False
                return
            
            self.changing_select_state = True

            if self._portgrp_widget is not None:
                if self._portgrp_widget.mouse_releasing:
                    self.setSelected(self._portgrp_widget.isSelected())
                elif not self._portgrp_widget.changing_select_state:
                    self._portgrp_widget.ensure_selection_with_ports()
            
            if self._lines_widgets:
                for line in self._lines_widgets:
                    line.update_line_gradient()
                    if self.isSelected():
                        line.setZValue(canvas.last_z_value)
                canvas.last_z_value += 1

            self.changing_select_state = False

        return QGraphicsItem.itemChange(self, change, value)

    def contextMenuEvent(self, event):
        if canvas.scene.get_zoom_scale() <= 0.4:
            # prefer move box if zoom is too low
            event.ignore()
            return
        
        if canvas.is_line_mov:
            return

        event.accept()
        canvas.scene.clearSelection()
        self.setSelected(True)

        canvas.menu_shown = True
        is_only_connect = bool(
            QApplication.keyboardModifiers() & Qt.ControlModifier)
        
        if is_only_connect:
            menu = ConnectMenu(self._port)
        else:
            menu = ConnectableContextMenu(self._port)

        act_x_sep_1 = menu.addSeparator()

        if (not is_only_connect
                and self._port_type == PortType.AUDIO_JACK
                and not self._port_subtype is PortSubType.CV
                and not self._portgrp_id):
            stereo_menu = QMenu(_translate('patchbay', "Set as Stereo with"), menu)
            menu.addMenu(stereo_menu)

            # get list of available mono ports settables as stereo with port
            port_id_cousin_list = list[int]()

            for port in canvas.list_ports(group_id=self._group_id):
                if (port.port_mode is self._port_mode
                        and port.port_type is self._port_type
                        and port.port_subtype is self._port_subtype):
                    port_id_cousin_list.append(port.port_id)

            selfport_index = port_id_cousin_list.index(self._port_id)
            at_least_one = False

            if selfport_index > 0:
                previous_port_id = port_id_cousin_list[selfport_index - 1]
                previous_port = canvas.get_port(self._group_id, previous_port_id)
                if previous_port and not previous_port.portgrp_id:
                    pp_action = stereo_menu.addAction(previous_port.port_name)
                    pp_action.setData((previous_port, self._port))
                    pp_action.triggered.connect(canvas.qobject.portgroup_the_ports)
                    at_least_one = True
                
            if selfport_index < len(port_id_cousin_list) - 1:
                next_port_id = port_id_cousin_list[selfport_index + 1]
                next_port = canvas.get_port(self._group_id, next_port_id)
                if next_port and not next_port.portgrp_id:
                    np_action = stereo_menu.addAction(next_port.port_name)
                    np_action.setData((self._port, next_port))
                    np_action.triggered.connect(canvas.qobject.portgroup_the_ports)
                    at_least_one = True

            if not at_least_one:
                act_x_setasstereo = stereo_menu.addAction('no available mono port')
                act_x_setasstereo.setEnabled(False)

        act_x_info = menu.addAction(_translate('patchbay', "Get &Info"))
        act_x_info.setIcon(QIcon.fromTheme('dialog-information'))
        act_x_rename = menu.addAction(_translate('patchbay', "&Rename"))

        if not features.port_info:
            act_x_info.setVisible(False)

        if not features.port_rename:
            act_x_rename.setVisible(False)

        if not (features.port_info and features.port_rename):
            act_x_sep_1.setVisible(False)

        if is_only_connect:
            act_x_info.setVisible(False)
            act_x_rename.setVisible(False)
            act_x_sep_1.setVisible(False)

        # prevent a bug that moves the box on mouse click after click to 
        # quit the menu if the two clic are at same point.
        self.parentItem().setFlag(QGraphicsItem.ItemIsMovable, False)

        # precise the menu start point to still view the port
        # and be able to read its portgroup name.        
        menu.show()
        
        start_point = canvas.scene.screen_position(
            self.scenePos() + QPointF(0.0, self._port_height))
        
        if (self._portgrp_id and self._port_mode is PortMode.INPUT
                and self._pg_pos + 1 <= self._pg_len // 2):
            start_point = canvas.scene.screen_position(
                self.scenePos() + QPointF(
                    0.0, self._port_height * (0.5 + self._pg_len / 2.0)))
            
        bottom_screen = QApplication.desktop().screenGeometry().bottom()
        more = 12 if self._port_mode is PortMode.OUTPUT else 0

        if start_point.y() + menu.height() > bottom_screen:
            start_point = canvas.scene.screen_position(
                self.scenePos() + QPointF(self._port_width + more, self._port_height))
        
        act_selected = menu.exec(start_point)
        
        if act_selected is act_x_info:
            canvas.callback(CallbackAct.PORT_INFO, self._group_id, self._port_id)

        elif act_selected is act_x_rename:
            canvas.callback(CallbackAct.PORT_RENAME, self._group_id, self._port_id)
            
        if act_selected is None:
            canvas.menu_click_pos = QCursor.pos()
        else:
            self.parentItem().setFlag(QGraphicsItem.ItemIsMovable, True)

    def boundingRect(self):
        if self._portgrp_id:
            if self._port_mode is PortMode.INPUT:
                return QRectF(0, 0, self._port_width, self._port_height)
            else:
                return QRectF(12, 0,
                              self._port_width, self._port_height)
        else:
            return QRectF(0, 0, self._port_width + 12, self._port_height)

    def paint(self, painter, option, widget):
        if canvas.loading_items:
            return
        
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing, True)
    
        theme = canvas.theme.port

        if self._port_type is PortType.AUDIO_JACK:
            if self._port_subtype is PortSubType.CV:
                theme = theme.cv
            else:
                theme = theme.audio
        elif self._port_type is PortType.MIDI_JACK:
            theme = theme.midi
        elif self._port_type is PortType.MIDI_ALSA:
            theme = theme.alsa
        elif self._port_type is PortType.VIDEO:
            theme = theme.video
        
        if self.isSelected():
            theme = theme.selected
        
        poly_image = theme.background_image()
        poly_color = theme.background_color()
        poly_color_alter = theme.background2_color()
        poly_pen = theme.fill_pen()
        #poly_pen.setJoinStyle(Qt.RoundJoin)
        text_pen = QPen(theme.text_color())

        # To prevent quality worsening
        poly_pen = QPen(poly_pen)
        poly_pen.setWidthF(poly_pen.widthF() + 0.00001)

        line_hinting = poly_pen.widthF() / 2
        p_height = canvas.theme.port_height
        text_y_pos = ((p_height - 0.667 * self._port_font.pixelSize()) / 2
                      + self._port_font.pixelSize() * 0.667)

        poly_locx = [0, 0, 0, 0, 0, 0]
        poly_corner_xhinting = ((float(canvas.theme.port_height)/2)
                                % floor(float(canvas.theme.port_height)/2))

        if poly_corner_xhinting == 0.0:
            poly_corner_xhinting = 0.5 * (1 - 7 / (float(canvas.theme.port_height)/2))

        is_cv_port = bool(self._port_subtype is PortSubType.CV)

        if self._port_mode is PortMode.INPUT:
            text_pos = QPointF(3, text_y_pos)

            if is_cv_port:
                poly_locx[0] = line_hinting
                poly_locx[1] = self._port_width + 7 - line_hinting
                poly_locx[2] = self._port_width + 7 - line_hinting
                poly_locx[3] = self._port_width + 7 - line_hinting
                poly_locx[4] = line_hinting
                poly_locx[5] = self._port_width

            elif self._port_type is PortType.MIDI_ALSA:
                poly_locx[0] = line_hinting
                poly_locx[1] = self._port_width + 5 - line_hinting
                poly_locx[2] = self._port_width + 8.5 - poly_corner_xhinting
                poly_locx[3] = self._port_width + 5 - line_hinting
                poly_locx[4] = line_hinting
                poly_locx[5] = self._port_width
                
            else:
                poly_locx[0] = line_hinting
                poly_locx[1] = self._port_width + 5 - line_hinting
                poly_locx[2] = self._port_width + 12 - poly_corner_xhinting
                poly_locx[3] = self._port_width + 5 - line_hinting
                poly_locx[4] = line_hinting
                poly_locx[5] = self._port_width

        elif self._port_mode is PortMode.OUTPUT:
            text_pos = QPointF(9, text_y_pos)

            if is_cv_port:
                poly_locx[0] = self._port_width + 12 - line_hinting
                poly_locx[1] = 5 + line_hinting
                poly_locx[2] = 5 + line_hinting
                poly_locx[3] = 5 + line_hinting
                poly_locx[4] = self._port_width + 12 - line_hinting
                poly_locx[5] = 12 - line_hinting
                
            elif self._port_type is PortType.MIDI_ALSA:
                poly_locx[0] = self._port_width + 12 - line_hinting
                poly_locx[1] = 7 + line_hinting
                poly_locx[2] = 3.5 + poly_corner_xhinting
                poly_locx[3] = 7 + line_hinting
                poly_locx[4] = self._port_width + 12 - line_hinting
                poly_locx[5] = 12 - line_hinting
                
            else:
                poly_locx[0] = self._port_width + 12 - line_hinting
                poly_locx[1] = 7 + line_hinting
                poly_locx[2] = 0 + poly_corner_xhinting
                poly_locx[3] = 7 + line_hinting
                poly_locx[4] = self._port_width + 12 - line_hinting
                poly_locx[5] = 12 - line_hinting

        else:
            self._logger.critical(
                f"paint() - invalid port mode {str(self._port_mode)}")
            return

        polygon = QPolygonF()

        if self._portgrp_id:
            first_of_portgrp = bool(self._pg_pos == 0)
            last_of_portgrp = bool(self._pg_pos + 1 == self._pg_len)

            if first_of_portgrp:
                polygon += QPointF(poly_locx[0] , line_hinting)
                polygon += QPointF(poly_locx[5] , line_hinting)
            else:
                polygon += QPointF(poly_locx[0] , 0)
                polygon += QPointF(poly_locx[5] , 0)

            if last_of_portgrp:
                polygon += QPointF(poly_locx[5], canvas.theme.port_height - line_hinting)
                polygon += QPointF(poly_locx[0], canvas.theme.port_height - line_hinting)
            else:
                polygon += QPointF(poly_locx[5], canvas.theme.port_height)
                polygon += QPointF(poly_locx[0], canvas.theme.port_height)

        elif self._port_type is PortType.MIDI_ALSA:
            polygon += QPointF(poly_locx[0], line_hinting)
            polygon += QPointF(poly_locx[2], line_hinting)
            polygon += QPointF(poly_locx[2], canvas.theme.port_height - line_hinting)
            polygon += QPointF(poly_locx[0], canvas.theme.port_height - line_hinting)
            
        else:
            polygon += QPointF(poly_locx[0], line_hinting)
            polygon += QPointF(poly_locx[1], line_hinting)
            polygon += QPointF(poly_locx[2], float(canvas.theme.port_height)/2)
            polygon += QPointF(poly_locx[3], canvas.theme.port_height - line_hinting)
            polygon += QPointF(poly_locx[4], canvas.theme.port_height - line_hinting)
            polygon += QPointF(poly_locx[0], line_hinting)

        if poly_image is not None:
            painter.setPen(Qt.NoPen)
            painter.setBrush(QBrush(poly_image))
            painter.drawPolygon(polygon)

        if poly_color_alter is not None:
            port_gradient = QLinearGradient(0, 0, 0, self._port_height)

            port_gradient.setColorAt(0, poly_color)
            port_gradient.setColorAt(0.5, poly_color_alter)
            port_gradient.setColorAt(1, poly_color)

            painter.setBrush(port_gradient)
        else:
            painter.setBrush(poly_color)
        
        painter.setPen(poly_pen)
        painter.drawPolygon(polygon)

        if not self._portgrp_id:
            if self._port_subtype is PortSubType.CV:
                poly_pen.setWidthF(2.000001)
                painter.setPen(poly_pen)

                y_line = canvas.theme.port_height / 2.0
                if self._port_mode is PortMode.OUTPUT:
                    painter.drawLine(
                        QPointF(0.0, y_line),
                        QPointF(float(poly_locx[1]) - line_hinting, y_line))
                elif self._port_mode is PortMode.INPUT:
                    painter.drawLine(
                        QPointF(float(self._port_width + 7), y_line),
                        QPointF(float(self._port_width + 12), y_line))

            elif (self._port_subtype is PortSubType.A2J
                    or self._port_type is PortType.MIDI_ALSA):
                # draw the little circle for a2j (or MidiBridge) port
                poly_pen.setWidthF(1.000001)
                
                # we emulate a hole in the port, so we need the background
                # of the box.
                parent = self.parentItem()
                box_theme = parent.get_theme()
                if parent.isSelected():
                    box_theme = box_theme.selected

                scene_col = canvas.theme.scene_background_color
                box_bg_col = box_theme.background_color()
                ra = box_bg_col.alphaF()
                rb = 1.0 - ra
                
                circle_bg_col = QColor()
                circle_bg_col.setRgbF(
                    scene_col.redF() * rb + box_bg_col.redF() * ra,
                    scene_col.greenF() * rb + box_bg_col.greenF() * ra,
                    scene_col.blueF() * rb + box_bg_col.blueF() * ra)
                
                painter.setBrush(circle_bg_col)

                ellipse_x = poly_locx[1]
                
                if self._port_type is PortType.MIDI_ALSA:
                    if self._port_mode is PortMode.OUTPUT:
                        ellipse_x -= 4
                    else:
                        ellipse_x += 4
                else:
                    if self._port_mode is PortMode.OUTPUT:
                        ellipse_x -= 2
                    elif self._port_mode is PortMode.INPUT:
                        ellipse_x += 2

                painter.drawEllipse(
                    QPointF(ellipse_x, canvas.theme.port_height / 2.0), 2.0, 2.0)

        painter.setPen(text_pen)
        painter.setFont(self._port_font)

        sizer = QFontMetrics(self._port_font)
        sep_width = sizer.width(self._trunck_sep)

        if self._portgrp_id:
            print_name_size = self.get_text_width()

            if self._port_mode is PortMode.OUTPUT:
                text_pos = QPointF(self._port_width + 9 - print_name_size,
                                   text_y_pos)

            if print_name_size > (self._port_width - 4):
                if poly_color_alter is not None:
                    painter.setPen(QPen(port_gradient, 3))
                else:
                    painter.setPen(QPen(poly_color, 3))
                painter.drawLine(
                    QPointF(float(poly_locx[5]), 3.0),
                    QPointF(float(poly_locx[5]),
                            float(canvas.theme.port_height - 3)))
                painter.setPen(text_pen)
                painter.setFont(self._port_font)

        painter.drawText(text_pos, self._print_name)
        
        if self._name_truncked:
            sep_x = text_pos.x() + sizer.width(self._print_name)
            
            painter.drawText(QPointF(sep_x + sep_width, text_pos.y()),
                             self._print_name_right)

            trunck_pen = QPen(text_pen)
            color = text_pen.color()
            color.setAlphaF(color.alphaF() * 0.25)
            trunck_pen.setColor(color)
            painter.setPen(trunck_pen)
            painter.drawText(QPointF(sep_x, text_pos.y() + 1), self._trunck_sep)

        painter.restore()




from pathlib import Path
from typing import TYPE_CHECKING
from PyQt5.QtWidgets import (QDialog, QApplication, QInputDialog,
                             QMessageBox, QWidget, QFileDialog)
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtCore import Qt, QProcess, QSettings

from .patchcanvas import patchcanvas, xdg
from .patchcanvas.theme_manager import ThemeData
from .tools_widgets import is_dark_theme
from .ui.canvas_options import Ui_CanvasOptions

if TYPE_CHECKING:
    from .patchbay_manager import PatchbayManager


_translate = QApplication.translate


class CanvasOptionsDialog(QDialog):    
    def __init__(self, parent: QWidget, manager: 'PatchbayManager',
                 settings: QSettings =None):
        QDialog.__init__(self, parent)
        self.ui = Ui_CanvasOptions()
        self.ui.setupUi(self)
        
        self.manager = manager
        sg = manager.sg
        self._settings = settings
        self._theme_changed = sg.theme_changed
        
        box_layout_dict = {
            1.0: _translate('box_layout', 'Choose the smallest area'),
            1.1: _translate('box_layout', 'Prefer large boxes'),
            1.4: _translate('box_layout', 'Almost only large boxes'),
            2.0: _translate('box_layout', 'Force large boxes')
        }
        
        for ratio, text in box_layout_dict.items():
            self.ui.comboBoxBoxesAutoLayout.addItem(text, ratio)

        if settings is not None:
            self.ui.checkBoxGracefulNames.setChecked(
                settings.value('Canvas/use_graceful_names', True, type=bool))
            self.ui.checkBoxA2J.setChecked(
                settings.value('Canvas/group_a2j_ports', True, type=bool))
            self.ui.checkBoxAlsa.setChecked(
                settings.value('Canvas/alsa_midi_enabled', False, type=bool))
            self.ui.checkBoxShadows.setChecked(
                settings.value('Canvas/box_shadows', False, type=bool))
            self.ui.checkBoxAutoSelectItems.setChecked(
                settings.value('Canvas/auto_select_items', False, type=bool))
            self.ui.checkBoxElastic.setChecked(
                settings.value('Canvas/elastic', True, type=bool))
            self.ui.checkBoxBordersNavigation.setChecked(
                settings.value('Canvas/borders_navigation', True, type=bool))
            self.ui.checkBoxPreventOverlap.setChecked(
                settings.value('Canvas/prevent_overlap', True, type=bool))
            self.ui.spinBoxMaxPortWidth.setValue(
                settings.value('Canvas/max_port_width', 170, type=int))
            self.ui.spinBoxDefaultZoom.setValue(
                settings.value('Canvas/default_zoom', 100, type=int))
            
            layout_ratio = settings.value(
                'Canvas/grouped_box_auto_layout_ratio', 1.0, type=float)
            
            if layout_ratio <= 1.0:
                box_index = 0
            elif layout_ratio <= 1.3:
                box_index = 1
            elif layout_ratio < 2.0:
                box_index = 2
            else:
                box_index = 3
                
            self.ui.comboBoxBoxesAutoLayout.setCurrentIndex(box_index)

        self.ui.comboBoxTheme.activated.connect(self._theme_box_activated)
        self.ui.pushButtonEditTheme.clicked.connect(self._edit_theme)
        self.ui.pushButtonDuplicateTheme.clicked.connect(self._duplicate_theme)
        self.ui.pushButtonPatchichiExport.clicked.connect(self._export_to_patchichi)
        
        self._current_theme_ref = ''
        self._theme_list = list[ThemeData]()
        
        # connect checkboxs and spinbox signals to patchbays signals
        self.ui.checkBoxGracefulNames.stateChanged.connect(
            manager.sg.graceful_names_changed)
        self.ui.checkBoxA2J.stateChanged.connect(
            manager.sg.a2j_grouped_changed)
        self.ui.checkBoxAlsa.stateChanged.connect(
            manager.sg.alsa_midi_enabled_changed)
        self.ui.checkBoxShadows.stateChanged.connect(
            manager.sg.group_shadows_changed)
        self.ui.checkBoxAutoSelectItems.stateChanged.connect(
            manager.sg.auto_select_items_changed)
        self.ui.checkBoxElastic.stateChanged.connect(
            manager.sg.elastic_changed)
        self.ui.checkBoxBordersNavigation.stateChanged.connect(
            manager.sg.borders_nav_changed)
        self.ui.checkBoxPreventOverlap.stateChanged.connect(
            manager.sg.prevent_overlap_changed)
        self.ui.spinBoxMaxPortWidth.valueChanged.connect(
            manager.sg.max_port_width_changed)
        self.ui.spinBoxDefaultZoom.valueChanged.connect(
            self._change_default_zoom)
        self.ui.comboBoxBoxesAutoLayout.currentIndexChanged.connect(
            self._grouped_box_auto_layout_changed)

    def _change_default_zoom(self, value: int):
        patchcanvas.set_default_zoom(value)
        patchcanvas.zoom_reset()

    def _theme_box_activated(self):
        current_theme_ref_id = self.ui.comboBoxTheme.currentData(Qt.UserRole)
        if current_theme_ref_id == self._current_theme_ref:
            return
        
        for theme_data in self._theme_list:
            if theme_data.ref_id == current_theme_ref_id:
                self.ui.pushButtonEditTheme.setEnabled(theme_data.editable)
                break

        self._theme_changed.emit(current_theme_ref_id)
        
    def _duplicate_theme(self):        
        new_theme_name, ok = QInputDialog.getText(
            self, _translate('patchbay_theme', 'New Theme Name'),
            _translate('patchbay_theme', 'Choose a name for the new theme :'))
        
        if not new_theme_name or not ok:
            return
        
        new_theme_name = new_theme_name.replace('/', '⁄')

        err = patchcanvas.copy_and_load_current_theme(new_theme_name)
        
        if err:
            message = _translate(
                'patchbay_theme', 'The copy of the theme directory failed')
            
            QMessageBox.warning(
                self, _translate('patchbay_theme', 'Copy failed !'), message)

    def _edit_theme(self):
        current_theme_ref_id = self.ui.comboBoxTheme.currentData(Qt.UserRole)
        
        for theme_data in self._theme_list:
            if (theme_data.ref_id == current_theme_ref_id
                    and theme_data.editable):
                # start the text editor process
                QProcess.startDetached('xdg-open', [theme_data.file_path])
                break

    def _export_to_patchichi(self):
        scenes_dir = xdg.xdg_data_home() / 'Patchichi' / 'scenes'

        if not scenes_dir.exists():
            try:
                scenes_dir.mkdir()
            except:
                pass
        
        if not scenes_dir.is_dir():
            scenes_dir = Path.home()
        
        ret, ok = QFileDialog.getSaveFileName(
            self,
            _translate('file_dialog', 'Where do you want to save this patchbay scene ?'),
            str(scenes_dir),
            _translate('file_dialog', 'Patchichi files (*.patchichi.json)'))

        if not ok:
            return
        
        self.manager.export_to_patchichi_json(Path(ret))

    def set_theme_list(self, theme_list: list[ThemeData]):
        self.ui.comboBoxTheme.clear()
        del self._theme_list
        self._theme_list = theme_list

        dark = '-dark' if is_dark_theme(self) else ''
        user_icon = QIcon(QPixmap(f':scalable/breeze{dark}/im-user'))

        for theme_data in theme_list:
            if theme_data.editable:
                self.ui.comboBoxTheme.addItem(
                    user_icon, theme_data.name, theme_data.ref_id)
            else:
                self.ui.comboBoxTheme.addItem(theme_data.name, theme_data.ref_id)

    def set_theme(self, theme_ref: str):
        for i in range(self.ui.comboBoxTheme.count()):
            ref_id = self.ui.comboBoxTheme.itemData(i, Qt.UserRole)
            if ref_id == theme_ref:
                self.ui.comboBoxTheme.setCurrentIndex(i)
                break
        else:
            # the new theme has not been found
            # update the list and select it if it exists
            self.set_theme_list(patchcanvas.list_themes())
            for i in range(self.ui.comboBoxTheme.count()):
                ref_id = self.ui.comboBoxTheme.itemData(i, Qt.UserRole)
                if ref_id == theme_ref:
                    self.ui.comboBoxTheme.setCurrentIndex(i)
                    
                    # update the edit button enable state
                    for theme_data in self._theme_list:
                        if theme_data.ref_id == ref_id:
                            self.ui.pushButtonEditTheme.setEnabled(
                                theme_data.editable)
                            break
                    break

    def enable_alsa_midi(self, yesno: bool):
        self.ui.checkBoxAlsa.setEnabled(yesno)
        if yesno:
            self.ui.checkBoxAlsa.setToolTip('')
        else:
            self.ui.checkBoxAlsa.setToolTip(
                _translate('alsa_midi', 
                           "ALSA python lib version is not present or too old.\n"
                           "Ensure to have python3-pyalsa >= 1.2.4"))

    def _grouped_box_auto_layout_changed(self, index: int):
        patchcanvas.set_grouped_box_layout_ratio(
            self.ui.comboBoxBoxesAutoLayout.currentData())

    def showEvent(self, event):
        self.set_theme_list(patchcanvas.list_themes())
        self.set_theme(patchcanvas.get_theme())
        QDialog.showEvent(self, event)

    def closeEvent(self, event):
        if self._settings is not None:
            self._settings.setValue('Canvas/use_graceful_names',
                                    self.ui.checkBoxGracefulNames.isChecked())
            self._settings.setValue('Canvas/group_a2j_ports',
                                    self.ui.checkBoxA2J.isChecked())
            self._settings.setValue('Canvas/alsa_midi_enabled',
                                    self.ui.checkBoxAlsa.isChecked())
            self._settings.setValue('Canvas/box_shadows',
                                    self.ui.checkBoxShadows.isChecked())
            self._settings.setValue('Canvas/auto_select_items',
                                    self.ui.checkBoxAutoSelectItems.isChecked())
            self._settings.setValue('Canvas/elastic',
                                    self.ui.checkBoxElastic.isChecked())
            self._settings.setValue('Canvas/borders_navigation',
                                    self.ui.checkBoxBordersNavigation.isChecked())
            self._settings.setValue('Canvas/prevent_overlap',
                                    self.ui.checkBoxPreventOverlap.isChecked())
            self._settings.setValue('Canvas/max_port_width',
                                    self.ui.spinBoxMaxPortWidth.value())
            self._settings.setValue('Canvas/default_zoom',
                                    self.ui.spinBoxDefaultZoom.value())
            self._settings.setValue('Canvas/theme',
                                    self.ui.comboBoxTheme.currentData())
            self._settings.setValue('Canvas/grouped_box_auto_layout_ratio',
                                    self.ui.comboBoxBoxesAutoLayout.currentData())
        QDialog.closeEvent(self, event)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys

import jacklib
from jacklib.helpers import get_jack_status_error_string


status = jacklib.jack_status_t()
client = jacklib.client_open("list-all-properties", jacklib.JackOptions.NO_START_SERVER, status)
err = get_jack_status_error_string(status)

if status.value:
    if status.value & jacklib.JackStatus.NAME_NOT_UNIQUE:
        print("Non-fatal JACK status: %s" % err, file=sys.stderr)
    elif status.value & jacklib.JackStatus.SERVER_STARTED:
        # Should not happen, since we use the JackOptions.NO_START_SERVER option
        print("Unexpected JACK status: %s" % err, file=sys.stderr)
    else:
        sys.exit("Error connecting to JACK server: %s" % err)

properties = jacklib.get_all_properties()

for subject, props in properties.items():
    print("Subject %s:" % subject)

    for prop in props:
        print("* {p.key}: {p.value} (type: {p.type})".format(p=prop))

    print('')

jacklib.client_close(client)



import pytest

import jacklib


@pytest.mark.jack_server_required
@pytest.mark.skipif(jacklib.jlib.jack_port_rename is None, reason="libjack does not implement 'jack_port_rename'.")
@pytest.mark.parametrize("porttype,flags,name", [
    (jacklib.JACK_DEFAULT_AUDIO_TYPE, jacklib.JackPortFlags.IS_OUTPUT, "output"),
    (jacklib.JACK_DEFAULT_AUDIO_TYPE, jacklib.JackPortFlags.IS_INPUT, "input"),
    (jacklib.JACK_DEFAULT_MIDI_TYPE, jacklib.JackPortFlags.IS_OUTPUT, "midi_output"),
    (jacklib.JACK_DEFAULT_MIDI_TYPE, jacklib.JackPortFlags.IS_INPUT, "midi_input"),
])
def test_port_rename(jack_client, porttype, flags, name):
    port = jacklib.port_register(jack_client, name, porttype, flags, 0)

    if not port:
        raise OSError(f"Could not register port {name}.")

    assert jacklib.port_short_name(port) == name
    res = jacklib.port_rename(jack_client, port, "renamed")
    assert res == 0
    assert jacklib.port_short_name(port) == "renamed"


@pytest.mark.jack_server_required
def test_port_rename_not_supported(jack_client):
    port = jacklib.port_register(jack_client, "output", jacklib.JACK_DEFAULT_AUDIO_TYPE, jacklib.JackPortFlags.IS_OUTPUT, 0)

    if not port:
        raise OSError(f"Could not register port {name}.")

    jacklib.jlib.jack_port_rename = None
    assert jacklib.port_short_name(port) == "output"

    res = jacklib.port_rename(jack_client, port, "renamed")
    assert res == -1
    assert jacklib.port_short_name(port) == "output"



"""JACK ctypes definitions for usage in python applications."""

# Copyright (C) 2010-2020 Filipe Coelho <falktx@falktx.com>
#               2016-2022 Christopher Arndt <info@chrisarndt.de>
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# For a full copy of the GNU General Public License see the file COPYING.md.

# -------------------------------------------------------------------------------------------------
# Imports (Global)

from collections import namedtuple
from ctypes import (
    pointer,
    POINTER,
    byref,
    c_char_p
)
from typing import Iterator

# import locals
from .cdll_funcs import get_jlib
from .enums import (
    JackOptions,
    JackStatus,
    JackLatencyCallbackMode,
    JackPortFlags,
    JackTransportState,
    JackPositionBits,
    JackSessionEvenType,
    JackSessionFlags,
    JackPropertyChange
)
from .types import (
    jack_uuid_t,
    jack_client_t,
    jack_port_t,
    jack_description_t,
    jack_session_command_t,
    JackThreadCallback,
    JackErrorCallback,
    JackTimebaseCallback
)
from .cb_setter import init_callback_setter, callback_setter

jlib = get_jlib()

# Defines
ENCODING = "utf-8"
JACK_MAX_FRAMES = 4294967295
JACK_LOAD_INIT_LIMIT = 1024
JACK_DEFAULT_AUDIO_TYPE = "32 bit float mono audio"
JACK_DEFAULT_MIDI_TYPE = "8 bit raw midi"

JACK_UUID_SIZE = 36
JACK_UUID_STRING_SIZE = JACK_UUID_SIZE + 1
JACK_UUID_EMPTY_INITIALIZER = 0

# Meta data
_JACK_METADATA_PREFIX = "http://jackaudio.org/metadata/"
JACK_METADATA_CONNECTED = _JACK_METADATA_PREFIX + "connected"
JACK_METADATA_EVENT_TYPES = _JACK_METADATA_PREFIX + "event-types"
JACK_METADATA_HARDWARE = _JACK_METADATA_PREFIX + "hardware"
JACK_METADATA_ICON_LARGE = _JACK_METADATA_PREFIX + "icon-large"
JACK_METADATA_ICON_NAME = _JACK_METADATA_PREFIX + "icon-name"
JACK_METADATA_ICON_SMALL = _JACK_METADATA_PREFIX + "icon-small"
JACK_METADATA_ORDER = _JACK_METADATA_PREFIX + "order"
JACK_METADATA_PORT_GROUP = _JACK_METADATA_PREFIX + "port-group"
JACK_METADATA_PRETTY_NAME = _JACK_METADATA_PREFIX + "pretty-name"
JACK_METADATA_SIGNAL_TYPE = _JACK_METADATA_PREFIX + "signal-type"


# Helper functions
def _e(s: str, encoding=ENCODING) -> bytes:
    if encoding:
        return s.encode(encoding)
    return s


def _d(s: bytes, encoding=ENCODING) -> str:
    if encoding:
        return s.decode(encoding)
    return s

init_callback_setter(jlib)

# JACK2 only
def get_version_string() -> str:
    if jlib.jack_get_version_string:
        return _d(jlib.jack_get_version_string())

    return ''

def client_open(client_name: str, options: JackOptions,
                status: JackStatus, uuid="") -> 'pointer[jack_client_t]':
    if jlib.jack_client_open:
        return jlib.jack_client_open(
            _e(client_name), options, status,
            _e(uuid) if uuid else None)

    return None

def client_rename(client: 'pointer[jack_client_t]', new_name: str):
    if jlib.jack_client_rename:
        return jlib.jack_client_rename(client, _e(new_name))

    return None

def client_close(client: 'pointer[jack_client_t]') -> int:
    if jlib.jack_client_close:
        return jlib.jack_client_close(client)

    return -1

def client_name_size() -> int:
    if jlib.jack_client_name_size:
        return jlib.jack_client_name_size()

    return 0

def get_client_name(client: 'pointer[jack_client_t]'):
    if jlib.jack_get_client_name:
        return jlib.jack_get_client_name(client)

    return None

def activate(client: 'pointer[jack_client_t]') -> int:
    if jlib.jack_activate:
        return jlib.jack_activate(client)

    return -1

def deactivate(client: 'pointer[jack_client_t]') -> int:
    if jlib.jack_deactivate:
        return jlib.jack_deactivate(client)

    return -1

# JACK2 only
def get_client_pid(name: str) -> int:
    if jlib.jack_get_client_pid:
        return jlib.jack_get_client_pid(_e(name))

    return 0

def is_realtime(client: 'pointer[jack_client_t]') -> int:
    if jlib.jack_is_realtime:
        return jlib.jack_is_realtime(client)

    return 0

# Non-Callback API
_thread_callback = None

def cycle_wait(client: 'pointer[jack_client_t]'):
    if jlib.jack_cycle_wait:
        return jlib.jack_cycle_wait(client)

    return 0

def cycle_signal(client: 'pointer[jack_client_t]', status: JackStatus):
    if jlib.jack_cycle_signal:
        jlib.jack_cycle_signal(client, status.value)

def set_process_thread(client: 'pointer[jack_client_t]', thread_callback, arg):
    if jlib.jack_set_process_thread:
        global _thread_callback
        _thread_callback = JackThreadCallback(thread_callback)
        return jlib.jack_set_process_thread(client, _thread_callback, arg)

    return -1

@callback_setter
def set_thread_init_callback(client, thread_init_callback, arg) -> int:
    ...

@callback_setter
def on_shutdown(client, shutdown_callback, arg):
    ...

@callback_setter
def on_info_shutdown(client, info_shutdown_callback, arg):
    ...

@callback_setter
def set_process_callback(client, process_callback, arg) -> int:
    ...

@callback_setter
def set_freewheel_callback(client, freewheel_callback, arg) -> int:
    ...

@callback_setter
def set_buffer_size_callback(client, bufsize_callback, arg) -> int:
    ...

@callback_setter
def set_sample_rate_callback(client, srate_callback, arg) -> int:
    ...

@callback_setter
def set_client_registration_callback(client, client_registration_callback, arg) -> int:
    ...

# JACK2 only
@callback_setter
def set_client_rename_callback(client, client_rename_callback, arg) -> int:
    ...

@callback_setter
def set_port_registration_callback(client, port_registration_callback, arg) -> int:
    ...

@callback_setter
def set_port_connect_callback(client, connect_callback, arg) -> int:
    ...

# JACK2 only
@callback_setter
def set_port_rename_callback(client, rename_callback, arg) -> int:
    ...

@callback_setter
def set_graph_order_callback(client, graph_callback, arg) -> int:
    ...

@callback_setter
def set_xrun_callback(client, xrun_callback, arg) -> int:
    ...

@callback_setter
def set_latency_callback(client, latency_callback, arg) -> int:
    ...

# -------------------------------------------------------------------------------------------------
# Server Control
def set_freewheel(client, onoff):
    return jlib.jack_set_freewheel(client, onoff)

def set_buffer_size(client, nframes):
    return jlib.jack_set_buffer_size(client, nframes)

def get_sample_rate(client):
    return jlib.jack_get_sample_rate(client)

def get_buffer_size(client):
    return jlib.jack_get_buffer_size(client)

def engine_takeover_timebase(client):
    return jlib.jack_engine_takeover_timebase(client)

def cpu_load(client):
    return jlib.jack_cpu_load(client)

# -------------------------------------------------------------------------------------------------
# Port Functions
def port_register(client, port_name, port_type, flags, buffer_size):
    return jlib.jack_port_register(
        client, _e(port_name), _e(port_type), flags, buffer_size)

def port_unregister(client, port):
    return jlib.jack_port_unregister(client, port)

def port_get_buffer(port, nframes):
    return jlib.jack_port_get_buffer(port, nframes)

def port_name(port) -> str:
    return _d(jlib.jack_port_name(port))

def port_short_name(port) -> str:
    return _d(jlib.jack_port_short_name(port))

def port_flags(port) -> int:
    return jlib.jack_port_flags(port)

def port_type(port) -> str:
    return _d(jlib.jack_port_type(port))

# JACK2 only
def port_type_id(port):
    return jlib.jack_port_type_id(port)

def port_is_mine(client, port):
    return jlib.jack_port_is_mine(client, port)

def port_connected(port):
    return jlib.jack_port_connected(port)

def port_connected_to(port, port_name):
    return jlib.jack_port_connected_to(port, _e(port_name))

def port_get_connections(port) -> Iterator[str]:
    ports = jlib.jack_port_get_connections(port)
    if not ports:
        return
    for port_name in ports:
        if port_name is None:
            break
        yield _d(port_name)

def port_get_all_connections(client, port) -> Iterator[str]:
    ports = jlib.jack_port_get_all_connections(client, port)
    if not ports:
        return
    for port_name in ports:
        if port_name is None:
            break
        yield _d(port_name)

def port_tie(src, dst):
    return jlib.jack_port_tie(src, dst)

def port_untie(port):
    return jlib.jack_port_untie(port)

def port_set_name(port, port_name: str):
    return jlib.jack_port_set_name(port, _e(port_name))

# JACK1 >= 0.125.0, JACK2 >= 1.19.11
def port_rename(client, port, port_name: str):
    if jlib.jack_port_rename:
        return jlib.jack_port_rename(client, port, _e(port_name))

    return -1

def port_set_alias(port, alias):
    return jlib.jack_port_set_alias(port, _e(alias))

def port_unset_alias(port, alias):
    return jlib.jack_port_unset_alias(port, _e(alias))

def port_get_aliases(port):
    # NOTE - this function has no 2nd argument in jacklib
    # Instead, aliases will be passed in return value, in form of (int ret, str alias1, str alias2)
    name_size = port_name_size()
    alias_type = c_char_p * 2
    aliases = alias_type(b" " * name_size, b" " * name_size)
    ret = jlib.jack_port_get_aliases(port, pointer(aliases))
    return ret, _d(aliases[0]), _d(aliases[1])

def port_request_monitor(port, onoff):
    return jlib.jack_port_request_monitor(port, onoff)

def port_request_monitor_by_name(client, port_name, onoff):
    return jlib.jack_port_request_monitor_by_name(client, _e(port_name), onoff)

def port_ensure_monitor(port, onoff):
    return jlib.jack_port_ensure_monitor(port, onoff)

def port_monitoring_input(port):
    return jlib.jack_port_monitoring_input(port)

def connect(client, source_port: str, destination_port: str):
    return jlib.jack_connect(client, _e(source_port), _e(destination_port))

def disconnect(client, source_port: str, destination_port: str):
    return jlib.jack_disconnect(client, _e(source_port), _e(destination_port))

def port_disconnect(client, port):
    return jlib.jack_port_disconnect(client, port)

def port_name_size():
    return jlib.jack_port_name_size()

def port_type_size():
    return jlib.jack_port_type_size()

def port_type_get_buffer_size(client, port_type):
    if jlib.jack_port_type_get_buffer_size:
        return jlib.jack_port_type_get_buffer_size(client, _e(port_type))

    return 0

def port_uuid(port):
    if jlib.jack_port_uuid:
        return jlib.jack_port_uuid(port)

    return -1


# -------------------------------------------------------------------------------------------------
# Latency Functions
def port_set_latency(port, nframes):
    jlib.jack_port_set_latency(port, nframes)

def port_get_latency_range(port, mode, range_):
    if jlib.jack_port_get_latency_range:
        jlib.jack_port_get_latency_range(port, mode, range_)

def port_set_latency_range(port, mode, range_):
    if jlib.jack_port_set_latency_range:
        jlib.jack_port_set_latency_range(port, mode, range_)

def recompute_total_latencies():
    return jlib.jack_recompute_total_latencies()

def port_get_latency(port):
    return jlib.jack_port_get_latency(port)

def port_get_total_latency(client, port):
    return jlib.jack_port_get_total_latency(client, port)

def recompute_total_latency(client, port):
    return jlib.jack_recompute_total_latency(client, port)

# -------------------------------------------------------------------------------------------------
# Port Searching
def get_ports(client, port_name_pattern=None,
              type_name_pattern=None, flags=0) -> 'pointer[c_char_p]':
    return jlib.jack_get_ports(
        client, _e(port_name_pattern or ""), _e(type_name_pattern or ""), flags
    )

def port_by_name(client, port_name: str) -> 'pointer[jack_port_t]':
    return jlib.jack_port_by_name(client, _e(port_name))

def port_by_id(client, port_id) -> 'pointer[jack_port_t]':
    return jlib.jack_port_by_id(client, port_id)

# -------------------------------------------------------------------------------------------------
# Time Functions
def frames_since_cycle_start(client):
    return jlib.jack_frames_since_cycle_start(client)

def frame_time(client):
    return jlib.jack_frame_time(client)

def last_frame_time(client):
    return jlib.jack_last_frame_time(client)

def get_cycle_times(client, current_frames, current_usecs, next_usecs, period_usecs):
    # JACK_OPTIONAL_WEAK_EXPORT
    if jlib.jack_frames_to_time:
        return jlib.jack_get_cycle_times(
            client, current_frames, current_usecs, next_usecs, period_usecs
        )
    return -1

def frames_to_time(client, nframes):
    return jlib.jack_frames_to_time(client, nframes)

def time_to_frames(client, time):
    return jlib.jack_time_to_frames(client, time)

def get_time():
    return jlib.jack_get_time()

# -------------------------------------------------------------------------------------------------
# Error Output

# TODO

# -------------------------------------------------------------------------------------------------
# Misc
_error_callback = None

def set_error_function(error_callback):
    global _error_callback
    if jlib.jack_set_error_function:
        _error_callback = JackErrorCallback(error_callback)
        jlib.jack_set_error_function(_error_callback)


def free(ptr):
    return jlib.jack_free(ptr)


# -------------------------------------------------------------------------------------------------
# Transport

_sync_callback = _timebase_callback = None

def release_timebase(client):
    return jlib.jack_release_timebase(client)

@callback_setter
def set_sync_callback(client, sync_callback, arg):
    ...

def set_sync_timeout(client, timeout):
    return jlib.jack_set_sync_timeout(client, timeout)

def set_timebase_callback(client, conditional, timebase_callback, arg):
    global _timebase_callback
    _timebase_callback = JackTimebaseCallback(timebase_callback)
    return jlib.jack_set_timebase_callback(
        client, conditional, _timebase_callback, arg)

def transport_locate(client, frame):
    return jlib.jack_transport_locate(client, frame)

def transport_query(client, pos):
    return jlib.jack_transport_query(client, pos)

def get_current_transport_frame(client):
    return jlib.jack_get_current_transport_frame(client)

def transport_reposition(client, pos):
    return jlib.jack_transport_reposition(client, pos)

def transport_start(client):
    return jlib.jack_transport_start(client)

def transport_stop(client):
    return jlib.jack_transport_stop(client)

# -------------------------------------------------------------------------------------------------
# MIDI
def midi_get_event_count(port_buffer):
    return jlib.jack_midi_get_event_count(port_buffer)

def midi_event_get(event, port_buffer, event_index):
    return jlib.jack_midi_event_get(event, port_buffer, event_index)

def midi_clear_buffer(port_buffer):
    return jlib.jack_midi_clear_buffer(port_buffer)

def midi_max_event_size(port_buffer):
    return jlib.jack_midi_max_event_size(port_buffer)

def midi_event_reserve(port_buffer, time, data_size):
    return jlib.jack_midi_event_reserve(port_buffer, time, data_size)

def midi_event_write(port_buffer, time, data, data_size):
    return jlib.jack_midi_event_write(port_buffer, time, data, data_size)

def midi_get_lost_event_count(port_buffer):
    return jlib.jack_midi_get_lost_event_count(port_buffer)

# -------------------------------------------------------------------------------------------------
# Session
@callback_setter
def set_session_callback(client, session_callback, arg):
    ...

def session_reply(client, event):
    if jlib.jack_session_reply:
        return jlib.jack_session_reply(client, event)

    return -1

def session_event_free(event):
    if jlib.jack_session_event_free:
        jlib.jack_session_event_free(event)

def client_get_uuid(client):
    if jlib.jack_client_get_uuid:
        return _d(jlib.jack_client_get_uuid(client))

    return None

def session_notify(client, target, type_, path):
    if jlib.jack_session_notify:
        return jlib.jack_session_notify(client, _e(target), type_, _e(path))

    return jack_session_command_t()

def session_commands_free(cmds):
    if jlib.jack_session_commands_free:
        jlib.jack_session_commands_free(cmds)

def get_uuid_for_client_name(client, client_name):
    if jlib.jack_get_uuid_for_client_name:
        return jlib.jack_get_uuid_for_client_name(client, _e(client_name))

    return None

def get_client_name_by_uuid(client, client_uuid):
    if jlib.jack_get_client_name_by_uuid:
        return jlib.jack_get_client_name_by_uuid(client, _e(client_uuid))

    return None

def reserve_client_name(client, name, uuid):
    if jlib.jack_reserve_client_name:
        return jlib.jack_reserve_client_name(client, _e(name), _e(uuid))

    return -1

def client_has_session_callback(client, client_name):
    if jlib.jack_client_has_session_callback:
        return jlib.jack_client_has_session_callback(client, _e(client_name))

    return -1

def uuid_parse(uuid_cstr):
    if jlib.jack_uuid_parse and uuid_cstr is not None:
        uuid = jack_uuid_t()
        res = jlib.jack_uuid_parse(uuid_cstr, byref(uuid))
        return uuid if res != -1 else None

    return -1

def uuid_unparse(uuid, encoding=ENCODING):
    if jlib.jack_uuid_unparse:
        uuid_str = c_char_p(b" " * JACK_UUID_STRING_SIZE)
        jlib.jack_uuid_unparse(uuid, uuid_str)
        return _d(uuid_str.value, encoding)

    return ""

# -------------------------------------------------------------------------------------------------
# Meta data

Property = namedtuple("Property", ("key", "value", "type"))

def free_description(description, free_description_itself=0):
    if jlib.jack_free_description:
        jlib.jack_free_description(description, free_description_itself)

def _decode_property(prop, encoding=ENCODING):
    key, value, type_ = prop.key, prop.data, prop.type
    decode_value = True

    try:
        key = _d(key, encoding)
    except UnicodeDecodeError:
        pass

    if type_:
        try:
            type_ = _d(type_, encoding)
        except UnicodeDecodeError:
            pass
        else:
            decode_value = type_.startswith("text/")

    if decode_value:
        try:
            value = _d(value, encoding)
        except UnicodeDecodeError:
            pass

    return Property(key, value, type_)

def get_all_properties(encoding=ENCODING):
    descriptions = POINTER(jack_description_t)()
    ret = jlib.jack_get_all_properties(byref(descriptions))
    results = {}

    if ret != -1:
        for p_idx in range(ret):
            description = descriptions[p_idx]

            if description.property_cnt:
                results[description.subject] = [
                    _decode_property(description.properties[p_idx], encoding)
                    for p_idx in range(description.property_cnt)
                ]

            free_description(description, 0)

    free(descriptions)
    return results

def get_properties(subject, encoding=ENCODING):
    description = jack_description_t()
    ret = jlib.jack_get_properties(subject, byref(description))
    results = []

    if ret != -1:
        for p_idx in range(description.property_cnt):
            results.append(_decode_property(description.properties[p_idx], encoding))

    free_description(byref(description), 0)
    return results

def get_client_properties(client, client_uuid, encoding=ENCODING):
    if isinstance(client_uuid, str):
        client_uuid = get_uuid_for_client_name(client, client_uuid)

    return get_properties(uuid_parse(client_uuid), encoding)

def get_port_properties(client, port, encoding=ENCODING):
    if not isinstance(port, POINTER(jack_port_t)):
        port = port_by_name(client, port)

    return get_properties(port_uuid(port), encoding)

def get_property(subject, key, encoding=ENCODING):
    # FIXME: how to handle non-null terminated data in value?
    #        We wouldn't know the length of the data in the value buffer.
    #        This seems to be an oversight in the JACK meta data API.
    value_c = c_char_p()
    type_c = c_char_p()
    ret = jlib.jack_get_property(subject, _e(key), byref(value_c), byref(type_c))
    value = value_c.value

    if ret != -1:
        decode_value = True

        if type_c:
            try:
                type_ = _d(type_c.value, encoding)
            except UnicodeDecodeError:
                # If type can't be decoded, we assume it's neither a mimetype
                # nor a URI, so we don't know how to interpret it and won't use
                # it to decide whether to decode the property value.
                type_ = type_c.value
            else:
                decode_value = type_.startswith("text/")

            free(type_c)
        else:
            type_ = None

        if decode_value:
            try:
                value = _d(value_c.value, encoding)
            except UnicodeDecodeError:
                pass

        free(value_c)
        return Property(key, value, type_)

def get_client_property(client, client_uuid, key, encoding=ENCODING):
    if isinstance(client_uuid, str):
        client_uuid = get_uuid_for_client_name(client, client_uuid)

    return get_property(uuid_parse(client_uuid), key, encoding)

def get_port_property(client, port, key, encoding=ENCODING):
    if not isinstance(port, POINTER(jack_port_t)):
        port = port_by_name(client, port)

    return get_property(port_uuid(port), key, encoding)

def get_port_pretty_name(client, port, encoding=ENCODING):
    prop = get_port_property(client, port, JACK_METADATA_PRETTY_NAME, encoding)
    return prop.value if prop else None

def remove_all_properties(client):
    return jlib.jack_remove_property(client)

def remove_properties(client, subject):
    return jlib.jack_remove_property(client, subject)

def remove_client_properties(client, client_uuid):
    if isinstance(client_uuid, str):
        client_uuid = get_uuid_for_client_name(client, client_uuid)

    return remove_properties(client, uuid_parse(client_uuid))

def remove_port_properties(client, port):
    if not isinstance(port, POINTER(jack_port_t)):
        port = port_by_name(client, port)

    return remove_properties(client, port_uuid(port))

def remove_property(client, subject, key, encoding=ENCODING):
    return jlib.jack_remove_property(client, subject, _e(key, encoding))

def remove_client_property(client, client_uuid, key, encoding=ENCODING):
    if isinstance(client_uuid, str):
        client_uuid = get_uuid_for_client_name(client, client_uuid)

    return remove_property(client, uuid_parse(client_uuid), key, encoding)

def remove_port_property(client, port, key, encoding=ENCODING):
    if not isinstance(port, POINTER(jack_port_t)):
        port = port_by_name(client, port)

    return remove_property(client, port_uuid(port), key, encoding)

def set_property(client, subject, key, value, type=None, encoding=ENCODING):
    if value is not None and encoding:
        value = _e(value, encoding)

    if type is not None and encoding:
        type = _e(type, encoding)

    return jlib.jack_set_property(client, subject, _e(key, encoding), value, type)

def set_client_property(client, client_uuid, key, value, type=None, encoding=ENCODING):
    if isinstance(client_uuid, str):
        client_uuid = get_uuid_for_client_name(client, client_uuid)

    uuid = uuid_parse(client_uuid)
    return set_property(client, uuid, key, value, type, encoding) if uuid != -1 else -1

def set_port_property(client, port, key, value, type=None, encoding=ENCODING):
    if not isinstance(port, POINTER(jack_port_t)):
        port = port_by_name(client, port)

    uuid = port_uuid(port)
    return set_property(client, uuid, key, value, type, encoding) if uuid != -1 else -1

def set_port_pretty_name(client, port, value, encoding=ENCODING):
    return set_port_property(client, port, JACK_METADATA_PRETTY_NAME, value, "text/plain", encoding)

@callback_setter
def set_property_change_callback(client, callback, arg=None):
    ...